from enum import Enum

from django.db.models import Avg, Count, Max, Min, Sum
from django.db.models.functions import (
    TruncDay,
    TruncMonth,
    TruncQuarter,
    TruncWeek,
    TruncYear,
)

REVERSE_FOREIGN_KEY_DELIMITER_COMMA = 1
REVERSE_FOREIGN_KEY_DELIMITER_SEMICOLON = 2
REVERSE_FOREIGN_KEY_DELIMITER_PIPE = 3
REVERSE_FOREIGN_KEY_DELIMITER_DASH = 4
REVERSE_FOREIGN_KEY_DELIMITER_SPACE = 5

REVERSE_FOREIGN_KEY_DELIMITER_CHOICES = [
    (REVERSE_FOREIGN_KEY_DELIMITER_COMMA, ', '),
    (REVERSE_FOREIGN_KEY_DELIMITER_SEMICOLON, ';'),
    (REVERSE_FOREIGN_KEY_DELIMITER_PIPE, ' | '),
    (REVERSE_FOREIGN_KEY_DELIMITER_DASH, ' - '),
    (REVERSE_FOREIGN_KEY_DELIMITER_SPACE, 'Space'),
]

REVERSE_FOREIGN_KEY_DELIMITER_VALUES = {
    REVERSE_FOREIGN_KEY_DELIMITER_COMMA: ', ',
    REVERSE_FOREIGN_KEY_DELIMITER_SEMICOLON: ';',
    REVERSE_FOREIGN_KEY_DELIMITER_PIPE: ' | ',
    REVERSE_FOREIGN_KEY_DELIMITER_DASH: ' - ',
    REVERSE_FOREIGN_KEY_DELIMITER_SPACE: ' ',
}

ANNOTATION_VALUE_YEAR = 1
ANNOTATION_VALUE_QUARTER = 2
ANNOTATION_VALUE_MONTH = 3
ANNOTATION_VALUE_WEEK = 4
ANNOTATION_VALUE_DAY = 5
ANNOTATION_VALUE_FINANCIAL_QUARTER = 6  # not used yet

ANNOTATION_VALUE_CHOICES = [
    (ANNOTATION_VALUE_YEAR, 'Year'),
    (ANNOTATION_VALUE_QUARTER, 'Quarter'),
    # (ANNOTATION_VALUE_FINANCIAL_QUARTER, 'Financial Quarter'),
    (ANNOTATION_VALUE_MONTH, 'Month'),
    (ANNOTATION_VALUE_WEEK, 'Week'),
    (ANNOTATION_VALUE_DAY, 'Day'),
]

ANNOTATION_VALUE_FUNCTIONS = {
    ANNOTATION_VALUE_YEAR: TruncYear,
    ANNOTATION_VALUE_QUARTER: TruncQuarter,
    ANNOTATION_VALUE_MONTH: TruncMonth,
    ANNOTATION_VALUE_WEEK: TruncWeek,
    ANNOTATION_VALUE_DAY: TruncDay,
}

REVERSE_FOREIGN_KEY_ANNOTATION_BOOLEAN_XOR = 1
REVERSE_FOREIGN_KEY_ANNOTATION_BOOLEAN_AND = 2
REVERSE_FOREIGN_KEY_ANNOTATION_BOOLEAN_ARRAY = 3

REVERSE_FOREIGN_KEY_ANNOTATION_BOOLEAN_CHOICES = [
    (REVERSE_FOREIGN_KEY_ANNOTATION_BOOLEAN_XOR, 'XOR'),
    (REVERSE_FOREIGN_KEY_ANNOTATION_BOOLEAN_AND, 'And'),
    (REVERSE_FOREIGN_KEY_ANNOTATION_BOOLEAN_ARRAY, 'Array'),
]

REVERSE_FOREIGN_KEY_ANNOTATION_DATE_ARRAY = 1
REVERSE_FOREIGN_KEY_ANNOTATION_DATE_MIN = 2
REVERSE_FOREIGN_KEY_ANNOTATION_DATE_MAX = 3

REVERSE_FOREIGN_KEY_ANNOTATION_DATE_CHOICES = [
    (REVERSE_FOREIGN_KEY_ANNOTATION_DATE_ARRAY, 'Array'),
    (REVERSE_FOREIGN_KEY_ANNOTATION_DATE_MIN, 'Min'),
    (REVERSE_FOREIGN_KEY_ANNOTATION_DATE_MAX, 'Max'),
]

ANNOTATION_CHART_SCALE = {
    ANNOTATION_VALUE_YEAR: 'year',
    ANNOTATION_VALUE_QUARTER: 'quarter',
    ANNOTATION_VALUE_MONTH: 'month',
    ANNOTATION_VALUE_WEEK: 'week',
    ANNOTATION_VALUE_DAY: 'day',
}

GENERATE_SERIES_INTERVALS = {
    ANNOTATION_VALUE_YEAR: '1 year',
    ANNOTATION_VALUE_QUARTER: '3 month',
    ANNOTATION_VALUE_MONTH: '1 month',
    ANNOTATION_VALUE_WEEK: '1 week',
    ANNOTATION_VALUE_DAY: '1 day',
}

ANNOTATION_CHOICE_NA = 0
ANNOTATION_CHOICE_SUM = 1
ANNOTATION_CHOICE_MAXIMUM = 2
ANNOTATION_CHOICE_MINIMUM = 3
ANNOTATION_CHOICE_COUNT = 4
ANNOTATION_CHOICE_AVERAGE_SUM_FROM_COUNT = 5

ANNOTATIONS_CHOICES = [
    (ANNOTATION_CHOICE_SUM, 'Sum'),
    (ANNOTATION_CHOICE_MAXIMUM, 'Maximum'),
    (ANNOTATION_CHOICE_MINIMUM, 'Minimum'),
    (ANNOTATION_CHOICE_COUNT, 'Count'),
    (ANNOTATION_CHOICE_AVERAGE_SUM_FROM_COUNT, 'Average Sum from Count'),
]

ANNOTATION_FUNCTIONS = {
    ANNOTATION_CHOICE_SUM: Sum,
    ANNOTATION_CHOICE_MAXIMUM: Max,
    ANNOTATION_CHOICE_MINIMUM: Min,
    ANNOTATION_CHOICE_COUNT: Count,
    ANNOTATION_CHOICE_AVERAGE_SUM_FROM_COUNT: Avg,
}

AGGREGATE_CLASSES = (Sum, Avg, Max, Min, Count)

ALIGNMENT_CHOICE_LEFT = 0
ALIGNMENT_CHOICE_CENTRE = 1
ALIGNMENT_CHOICE_RIGHT = 2

ALIGNMENT_CHOICES = [
    (ALIGNMENT_CHOICE_LEFT, 'Left'),
    (ALIGNMENT_CHOICE_CENTRE, 'Centre'),
    (ALIGNMENT_CHOICE_RIGHT, 'Right'),
]

ALIGNMENT_CLASS = {
    ALIGNMENT_CHOICE_LEFT: '',
    ALIGNMENT_CHOICE_CENTRE: 'dt-center',
    ALIGNMENT_CHOICE_RIGHT: 'dt-right',
}

DATE_FORMAT_TYPE_DD_MM_YY_SLASH = 1
DATE_FORMAT_TYPE_DD_MM_YYYY_SLASH = 2
DATE_FORMAT_TYPE_MM_DD_YY_SLASH = 3
DATE_FORMAT_TYPE_MM_DD_YYYY_SLASH = 4
DATE_FORMAT_TYPE_DD_MM_YY_DASH = 5
DATE_FORMAT_TYPE_DD_MM_YYYY_DASH = 6
DATE_FORMAT_TYPE_MM_DD_YY_DASH = 7
DATE_FORMAT_TYPE_MM_DD_YYYY_DASH = 8
DATE_FORMAT_TYPE_WORDS_MM_DD_YYYY = 9
DATE_FORMAT_TYPE_WORDS_DD_MM_YYYY = 10
DATE_FORMAT_TYPE_SHORT_WORDS_MM_DD_YYYY = 11
DATE_FORMAT_TYPE_SHORT_WORDS_DD_MM_YYYY = 12
DATE_FORMAT_TYPE_YYYY = 13
DATE_FORMAT_TYPE_YY = 14
DATE_FORMAT_TYPE_MM_YY = 15
DATE_FORMAT_TYPE_MM_YYYY = 16
DATE_FORMAT_TYPE_WORDS_MM_YY = 17
DATE_FORMAT_TYPE_WORDS_MM_YYYY = 18
DATE_FORMAT_TYPE_SHORT_WORDS_MM_YY = 19
DATE_FORMAT_TYPE_SHORT_WORDS_MM_YYYY = 20
DATE_FORMAT_TYPE_MM = 21
DATE_FORMAT_TYPE_WORDS_MM = 22
DATE_FORMAT_TYPE_SHORT_WORDS_MM = 23
DATE_FORMAT_TYPE_WW = 24
DATE_FORMAT_TYPE_WW_YYYY_DASH = 25
DATE_FORMAT_TYPE_WW_YY_DASH = 26
DATE_FORMAT_TYPE_YYYY_MM = 27


DATE_FORMAT_TYPES_DJANGO_FORMAT = {
    DATE_FORMAT_TYPE_DD_MM_YY_SLASH: '%d/%m/%y',
    DATE_FORMAT_TYPE_DD_MM_YYYY_SLASH: '%d/%m/%Y',
    DATE_FORMAT_TYPE_MM_DD_YY_SLASH: '%m/%d/%y',
    DATE_FORMAT_TYPE_MM_DD_YYYY_SLASH: '%m/%d/%Y',
    DATE_FORMAT_TYPE_DD_MM_YY_DASH: '%d-%m-%y',
    DATE_FORMAT_TYPE_DD_MM_YYYY_DASH: '%d-%m-%Y',
    DATE_FORMAT_TYPE_MM_DD_YY_DASH: '%m-%d-%y',
    DATE_FORMAT_TYPE_MM_DD_YYYY_DASH: '%m-%d-%Y',
    DATE_FORMAT_TYPE_WORDS_MM_DD_YYYY: '%B %d %Y',
    DATE_FORMAT_TYPE_WORDS_DD_MM_YYYY: '%d %B %Y',
    DATE_FORMAT_TYPE_SHORT_WORDS_MM_DD_YYYY: '%b %d %Y',
    DATE_FORMAT_TYPE_SHORT_WORDS_DD_MM_YYYY: '%d %b %Y',
    DATE_FORMAT_TYPE_YYYY: '%Y',
    DATE_FORMAT_TYPE_YY: '%y',
    DATE_FORMAT_TYPE_MM_YY: '%m %y',
    DATE_FORMAT_TYPE_YYYY_MM: '%Y %m',
    DATE_FORMAT_TYPE_MM_YYYY: '%m %Y',
    DATE_FORMAT_TYPE_WORDS_MM_YY: '%B %y',
    DATE_FORMAT_TYPE_WORDS_MM_YYYY: '%B %Y',
    DATE_FORMAT_TYPE_SHORT_WORDS_MM_YY: '%b %y',
    DATE_FORMAT_TYPE_SHORT_WORDS_MM_YYYY: '%b %Y',
    DATE_FORMAT_TYPE_MM: '%m',
    DATE_FORMAT_TYPE_WORDS_MM: '%B',
    DATE_FORMAT_TYPE_SHORT_WORDS_MM: '%b',
    DATE_FORMAT_TYPE_WW: '%W',
    DATE_FORMAT_TYPE_WW_YYYY_DASH: '%W-%Y',
    DATE_FORMAT_TYPE_WW_YY_DASH: '%W-%y',
}

DATE_FORMAT_TYPES = [
    (DATE_FORMAT_TYPE_DD_MM_YY_SLASH, 'dd/mm/yy'),
    (DATE_FORMAT_TYPE_DD_MM_YYYY_SLASH, 'dd/mm/yyyy'),
    (DATE_FORMAT_TYPE_MM_DD_YY_SLASH, 'mm/dd/yy'),
    (DATE_FORMAT_TYPE_MM_DD_YYYY_SLASH, 'mm/dd/yyyy'),
    (DATE_FORMAT_TYPE_DD_MM_YY_DASH, 'dd-mm-yy'),
    (DATE_FORMAT_TYPE_DD_MM_YYYY_DASH, 'dd-mm-yyyy'),
    (DATE_FORMAT_TYPE_MM_DD_YY_DASH, 'mm-dd-yy'),
    (DATE_FORMAT_TYPE_MM_DD_YYYY_DASH, 'mm-dd-yyyy'),
    (DATE_FORMAT_TYPE_WORDS_MM_DD_YYYY, 'MM d yyyy'),
    (DATE_FORMAT_TYPE_WORDS_DD_MM_YYYY, 'd MM yyyy'),
    (DATE_FORMAT_TYPE_SHORT_WORDS_MM_DD_YYYY, 'M d yyyy'),
    (DATE_FORMAT_TYPE_SHORT_WORDS_DD_MM_YYYY, 'd M yyyy'),
    (DATE_FORMAT_TYPE_YYYY, 'yyyy'),
    (DATE_FORMAT_TYPE_YY, 'yy'),
    (DATE_FORMAT_TYPE_MM_YY, 'mm yy'),
    (DATE_FORMAT_TYPE_MM_YYYY, 'mm yyyy'),
    (DATE_FORMAT_TYPE_YYYY_MM, 'yyyy mm'),
    (DATE_FORMAT_TYPE_WORDS_MM_YY, 'MM yy'),
    (DATE_FORMAT_TYPE_WORDS_MM_YYYY, 'MM yyyy'),
    (DATE_FORMAT_TYPE_SHORT_WORDS_MM_YY, 'b yy'),
    (DATE_FORMAT_TYPE_SHORT_WORDS_MM_YYYY, 'b yyyy'),
    (DATE_FORMAT_TYPE_MM, 'mm'),
    (DATE_FORMAT_TYPE_WORDS_MM, 'MM'),
    (DATE_FORMAT_TYPE_SHORT_WORDS_MM, 'd'),
    (DATE_FORMAT_TYPE_WW, 'WW'),
    (DATE_FORMAT_TYPE_WW_YYYY_DASH, 'WW-YYYY'),
    (DATE_FORMAT_TYPE_WW_YY_DASH, 'WW-YY'),
]

DISPLAY_OPTION_NONE = 0
DISPLAY_OPTION_1_PER_ROW = 1
DISPLAY_OPTION_2_PER_ROW = 2
DISPLAY_OPTION_3_PER_ROW = 3
DISPLAY_OPTION_4_PER_ROW = 4

DISPLAY_OPTION_CHOICES = [
    (DISPLAY_OPTION_NONE, 'None/Default'),
    (DISPLAY_OPTION_1_PER_ROW, '1 Report per Row'),
    (DISPLAY_OPTION_2_PER_ROW, '2 Reports per Row'),
    (DISPLAY_OPTION_3_PER_ROW, '3 Reports per Row'),
    (DISPLAY_OPTION_4_PER_ROW, '4 Reports per Row'),
]

DISPLAY_OPTION_CLASSES = {
    DISPLAY_OPTION_1_PER_ROW: ' col-12',
    DISPLAY_OPTION_2_PER_ROW: ' col-12 col-md-12 col-lg-6',
    DISPLAY_OPTION_3_PER_ROW: ' col-12 col-md-12 col-lg-4',
    DISPLAY_OPTION_4_PER_ROW: ' col-12 col-md-12 col-lg-3',
}

DEFAULT_DATE_FORMAT = {
    ANNOTATION_VALUE_YEAR: DATE_FORMAT_TYPE_YYYY,
    ANNOTATION_VALUE_QUARTER: DATE_FORMAT_TYPE_WORDS_MM_YY,
    ANNOTATION_VALUE_MONTH: DATE_FORMAT_TYPE_WORDS_MM_YY,
    ANNOTATION_VALUE_WEEK: DATE_FORMAT_TYPE_WW_YY_DASH,
    ANNOTATION_VALUE_DAY: DATE_FORMAT_TYPE_DD_MM_YY_SLASH,
}

CALENDAR_VIEW_TYPE_MONTH = 1
CALENDAR_VIEW_TYPE_GRID_WEEK = 2
CALENDAR_VIEW_TYPE_LIST_WEEK = 3
CALENDAR_VIEW_TYPE_DAY = 4
CALENDAR_VIEW_TYPE_YEAR = 5

CALENDAR_VIEW_TYPE_CHOICES = (
    (CALENDAR_VIEW_TYPE_MONTH, 'Month'),
    (CALENDAR_VIEW_TYPE_GRID_WEEK, 'Grid Week'),
    (CALENDAR_VIEW_TYPE_LIST_WEEK, 'List Week'),
    (CALENDAR_VIEW_TYPE_DAY, 'Day'),
    (CALENDAR_VIEW_TYPE_YEAR, 'Year'),
)


class FieldType(Enum):
    NULL_FIELD = 1
    ABSTRACT_USER = 2
    FILTER_FOREIGN_KEY = 3
    STRING = 4
    NUMBER = 5
    BOOLEAN = 6
    DATE = 7
    MANY_TO_MANY = 8



import base64
import copy
import json
import math
from datetime import datetime, timedelta

from date_offset.date_offset import DateOffset
from dateutil.relativedelta import relativedelta
from django.apps import apps
from django.core.exceptions import FieldDoesNotExist, FieldError
from django.db import DataError, ProgrammingError
from django.db.models import Q
from django.forms import ChoiceField
from django.shortcuts import get_object_or_404
from django.utils.safestring import mark_safe
from django.views.generic import TemplateView
from django_datatables.datatables import DatatableTable
from django_menus.menu import MenuItem
from django_modals.widgets.select2 import Select2

from advanced_report_builder.column_types import NUMBER_FIELDS
from advanced_report_builder.columns import ReportBuilderDateColumn
from advanced_report_builder.exceptions import ReportError
from advanced_report_builder.field_utils import ReportBuilderFieldUtils
from advanced_report_builder.globals import (
    ANNOTATION_VALUE_DAY,
    ANNOTATION_VALUE_FINANCIAL_QUARTER,
    ANNOTATION_VALUE_FUNCTIONS,
    ANNOTATION_VALUE_MONTH,
    ANNOTATION_VALUE_QUARTER,
    ANNOTATION_VALUE_WEEK,
    ANNOTATION_VALUE_YEAR,
)
from advanced_report_builder.models import ReportType
from advanced_report_builder.utils import (
    count_days,
    get_report_builder_class,
    split_attr,
    split_slug,
)
from advanced_report_builder.variable_date import VariableDate
from advanced_report_builder.views.helpers import QueryBuilderForm
from advanced_report_builder.views.report import ReportBase
from advanced_report_builder.views.report_utils_mixin import ReportUtilsMixin
from advanced_report_builder.views.targets.utils import get_target_value


class ChartJSTable(DatatableTable):
    def __init__(self, *args, **kwargs):
        pk = kwargs.pop('pk', None)
        self.axis_scale = kwargs.pop('axis_scale', None)
        self.targets = kwargs.pop('targets', None)
        self.raw_data = None

        super().__init__(*args, **kwargs)
        if pk:
            self.filter['pk'] = pk

    def model_table_setup(self):
        try:
            targets_data = []
            error = False
            try:
                data = self.get_table_array(self.kwargs.get('request'), self.get_query())
                self.raw_data = data
            except (DataError, FieldError):
                data = [['N/A']]
                error = True
            if self.targets is not None and not error:
                targets = self.targets.all()
                if targets and len(data) > 0:
                    targets_data = self.process_data_structure_target(targets=targets, data=data)
            return mark_safe(
                json.dumps(
                    {
                        'initsetup': json.loads(self.col_def_str()),
                        'data': data,
                        'targets': targets_data,
                        'row_titles': self.all_titles(),
                        'table_id': self.table_id,
                    }
                )
            )
        except (ProgrammingError, TypeError, ValueError, KeyError) as e:
            raise ReportError(e)

    def process_data_structure_target(self, targets, data):
        results = []
        for target in targets:
            new_data_structure = []
            for data_dict in data:
                target_value = self.process_target_results(data_dict=data_dict, target=target)
                new_data_structure.append(target_value)
            label = target.name + ' Target'
            colour = '#' + target.colour
            results.append(
                {
                    'label': label,
                    'data': new_data_structure,
                    'borderWidth': 1,
                    'backgroundColor': [colour for _ in range(len(new_data_structure))],
                    'borderColor': '#' + target.colour,
                }
            )
        return results

    @staticmethod
    def process_target_results(data_dict, target):
        """
        Get the correct target data from from the target.
        :param data_dict:
        :param target:
        :return:
        """
        start_date = datetime.strptime(data_dict[0], '%Y-%m-%d').date()

        date_offset = DateOffset()

        # since the targets are in months, this is the value we need.
        first_day_month = start_date.replace(day=1)
        next_date = date_offset.get_offset('1m', first_day_month) - timedelta(days=1)

        target_value = get_target_value(
            min_date=first_day_month,
            max_date=next_date,
            target=target,
            month_range=True,
        )

        return target_value


class ChartBaseView(ReportBase, ReportUtilsMixin, TemplateView):
    date_field = ReportBuilderDateColumn
    chart_js_table = ChartJSTable

    template_name = 'advanced_report_builder/charts/report.html'

    def __init__(self, *args, **kwargs):
        self.chart_report = None
        self.show_toolbar = False
        self.table = None
        super().__init__(*args, **kwargs)

    def dispatch(self, request, *args, **kwargs):
        self.slug = split_slug(kwargs.get('slug'))
        self.enable_edit = kwargs.get('enable_edit')
        self.dashboard_report = kwargs.get('dashboard_report')
        if self.enable_edit or (self.dashboard_report and not self.dashboard_report.top) or not self.dashboard_report:
            self.show_toolbar = True
        return super().dispatch(request, *args, **kwargs)

    def extra_filters(self, query):
        report_query = self.get_report_query(report=self.chart_report)
        if report_query:
            query = self.process_query_filters(query=query, search_filter_data=report_query.query)
        return query

    def get_date_format(self):
        return '%Y-%m-%d'

    def get_date_field(self, index, fields, base_model, table):
        field_name = self.chart_report.date_field
        if field_name is None:
            return
        report_builder_class = get_report_builder_class(model=base_model, report_type=self.chart_report.report_type)

        django_field, col_type_override, _, _ = self.get_field_details(
            base_model=base_model,
            field=field_name,
            report_builder_class=report_builder_class,
            table=table,
        )

        if col_type_override:
            field_name = col_type_override.field

        date_format = self.get_date_format()

        date_function_kwargs = {'title': field_name, 'date_format': date_format}

        annotations_value = self.chart_report.axis_scale

        new_field_name = f'{annotations_value}_{field_name}_{index}'
        function = ANNOTATION_VALUE_FUNCTIONS[annotations_value]
        date_function_kwargs['annotations_value'] = {new_field_name: function(field_name)}
        field_name = new_field_name

        date_function_kwargs.update({'field': field_name, 'column_name': field_name, 'model_path': ''})

        field = self.date_field(**date_function_kwargs)
        fields.append(field)
        return field_name

    @staticmethod
    def _set_multiple_title(database_values, value_prefix, fields, text):
        results = {}
        for field in fields:
            value = database_values[value_prefix + '__' + field]
            results[field] = value
        return text.format(**results)

    def process_query_results(self, base_model, table):
        fields = []
        date_field_name = self.get_date_field(0, fields, base_model=base_model, table=table)
        if date_field_name is not None:
            table.order_by = [date_field_name]
        else:
            table.order_by = ['id']

        if not self.chart_report.fields:
            return fields
        chart_fields = self.chart_report.fields

        report_builder_class = get_report_builder_class(model=base_model, report_type=self.chart_report.report_type)

        for index, table_field in enumerate(chart_fields, 1):
            field = table_field['field']

            django_field, col_type_override, _, _ = self.get_field_details(
                base_model=base_model,
                field=field,
                report_builder_class=report_builder_class,
                table=table,
            )

            if isinstance(django_field, NUMBER_FIELDS) or (
                col_type_override is not None and col_type_override.annotations
            ):
                data_attr = split_attr(table_field)
                if data_attr.get('multiple_columns') == '1':
                    query = self.extra_filters(query=table.model.objects)

                    multiple_column_field = data_attr.get('multiple_column_field')
                    if multiple_column_field is None:
                        return
                    report_builder_class = self._get_report_builder_class(
                        base_model=base_model,
                        field_str=multiple_column_field,
                        report_builder_class=report_builder_class,
                    )
                    try:
                        _fields = report_builder_class.default_multiple_column_fields
                    except AttributeError:
                        return
                    default_multiple_column_fields = [multiple_column_field + '__' + x for x in _fields]
                    order_by = f'{multiple_column_field}__{report_builder_class.default_multiple_pk}'
                    positive_colour_field = self.get_field_path(
                        field_name=data_attr.get('positive_colour_field'),
                        base_model=base_model,
                        report_builder_class=report_builder_class,
                        table=table,
                    )
                    if positive_colour_field is not None:
                        default_multiple_column_fields.append(positive_colour_field)

                    negative_colour_field = self.get_field_path(
                        field_name=data_attr.get('negative_colour_field'),
                        base_model=base_model,
                        report_builder_class=report_builder_class,
                        table=table,
                    )
                    if negative_colour_field is not None:
                        default_multiple_column_fields.append(negative_colour_field)

                    results = (
                        query.order_by(order_by)
                        .distinct(multiple_column_field)
                        .values(multiple_column_field, *default_multiple_column_fields)
                    )
                    held_kwargs = copy.deepcopy(col_type_override.kwargs)
                    for multiple_index, result in enumerate(results):
                        col_type_override.kwargs = copy.deepcopy(held_kwargs)

                        suffix = self._set_multiple_title(
                            database_values=result,
                            value_prefix=multiple_column_field,
                            fields=_fields,
                            text=report_builder_class.default_multiple_column_text,
                        )
                        extra_filter = Q((multiple_column_field, result[multiple_column_field]))

                        additional_options = {}
                        if positive_colour_field is not None:
                            additional_options['positive_colour'] = result[positive_colour_field]

                        if negative_colour_field is not None:
                            additional_options['negative_colour'] = result[negative_colour_field]

                        self.get_number_field(
                            annotations_type=self.chart_report.axis_value_type,
                            append_annotation_query=True,
                            index=f'{index}_{multiple_index}',
                            data_attr=data_attr,
                            table_field=table_field,
                            fields=fields,
                            col_type_override=col_type_override,
                            extra_filter=extra_filter,
                            title_suffix=suffix,
                            multiple_index=multiple_index,
                            additional_options=additional_options,
                        )

                else:
                    self.get_number_field(
                        annotations_type=self.chart_report.axis_value_type,
                        append_annotation_query=False,
                        index=index,
                        data_attr=data_attr,
                        table_field=table_field,
                        fields=fields,
                        col_type_override=col_type_override,
                    )
        return fields

    def get_field_path(self, field_name, base_model, report_builder_class, table):
        if field_name:
            _, _, _, field_path = self.get_field_details(
                base_model=base_model,
                field=field_name,
                report_builder_class=report_builder_class,
                table=table,
            )
            return field_path
        return None

    # noinspection PyUnresolvedReferences
    @staticmethod
    def add_colour_offset(colour, multiple_index):
        if multiple_index > 0:
            offset = 50 * multiple_index
            colour_list = list(int(colour[i : i + 2], 16) for i in (0, 2, 4))
            _, colour_list[0] = divmod(colour_list[0] + offset, 255)
            _, colour_list[1] = divmod(colour_list[1] + offset, 255)
            _, colour_list[2] = divmod(colour_list[2] + offset, 255)
            return '{:02x}{:02x}{:02x}'.format(*colour_list)
        else:
            return colour

    def setup_table(self, base_model):
        if hasattr(self.chart_report, 'axis_scale'):
            self.table = self.chart_js_table(model=base_model, axis_scale=self.chart_report.axis_scale)
        else:
            self.table = self.chart_js_table(model=base_model)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        base_model = self.chart_report.get_base_model()
        if base_model:
            self.setup_table(base_model=base_model)
            self.table.extra_filters = self.extra_filters
            try:
                fields = self.process_query_results(base_model=base_model, table=self.table)
            except (FieldError, FieldDoesNotExist) as e:
                raise ReportError(e)
            self.table.add_columns(*fields)
            context['datatable'] = self.table
        context['show_toolbar'] = self.show_toolbar
        context['title'] = self.get_title()
        return context

    def setup_menu(self):
        if not self.show_toolbar:
            return

        if self.dashboard_report and self.enable_edit:
            report_menu = self.pod_dashboard_edit_menu()
        elif self.dashboard_report and not self.enable_edit:
            report_menu = self.pod_dashboard_view_menu()
        else:
            report_menu = self.pod_report_menu()

        self.add_menu('button_menu', 'button_group').add_items(
            *report_menu,
            *self.queries_menu(report=self.report, dashboard_report=self.dashboard_report),
        )

    def pod_dashboard_edit_menu(self):
        return [
            MenuItem(
                f'advanced_report_builder:dashboard_report_modal,pk-{self.dashboard_report.id}',
                menu_display='Edit',
                font_awesome='fas fa-pencil-alt',
                css_classes=['btn-primary'],
            )
        ]

    # noinspection PyMethodMayBeStatic
    def pod_dashboard_view_menu(self):
        return []

    def pod_report_menu(self):
        query_id = self.slug.get(f'query{self.chart_report.id}')
        slug_str = ''
        if query_id:
            slug_str = f'-query_id-{query_id}'
        return self.edit_report_menu(
            request=self.request,
            chart_report_id=self.chart_report.id,
            slug_str=slug_str,
        )

    @staticmethod
    def edit_report_menu(request, chart_report_id, slug_str):
        return []

    def get_period_divider(
        self,
        annotation_value_choice,
        start_date_type,
        end_date_type,
        exclude_weekdays=None,
        exclude_dates=None,
        financial_year_start_month: int = 4,  # April default
    ):
        """
        Calculates the number of periods (year, quarter, month, week, day)
        between two variable dates.
        Supports working-day exclusions and financial quarters.

        financial_year_start_month:
            1 = Jan (default calendar year)
            4 = Apr (UK fiscal year)
            7 = Jul (AU fiscal year)
        """
        variable_date = VariableDate()

        financial_year_start_month = self.get_financial_month()

        start_date_and_time, _, _ = variable_date.get_variable_dates(
            range_type=start_date_type, financial_year_start_month=financial_year_start_month
        )
        _, end_date_and_time, _ = variable_date.get_variable_dates(
            range_type=end_date_type, financial_year_start_month=financial_year_start_month
        )

        start_date = start_date_and_time.date()
        end_date = end_date_and_time.date()

        # YEAR
        if annotation_value_choice == ANNOTATION_VALUE_YEAR:
            if exclude_weekdays or exclude_dates:
                total_working_days = count_days(
                    start_date,
                    end_date,
                    exclude_weekdays=exclude_weekdays,
                    exclude_dates=exclude_dates,
                )
                divider = max(1, math.ceil(total_working_days / 260))
            else:
                divider = abs(end_date.year - start_date.year) + 1

        # FINANCIAL QUARTER
        elif annotation_value_choice == ANNOTATION_VALUE_FINANCIAL_QUARTER:
            # Financial year offset (e.g. 4 → April)
            fy_offset = financial_year_start_month - 1

            # Convert actual months to financial-year-relative months (0–11)
            start_fm = (start_date.year * 12) + (start_date.month - 1) - fy_offset
            end_fm = (end_date.year * 12) + (end_date.month - 1) - fy_offset

            # Month difference (inclusive of partial months if days > 0)
            delta = relativedelta(end_date, start_date)

            # Adjust using financial-year-relative absolute month numbers
            adjusted_months = (end_fm - start_fm) + (1 if delta.days > 0 else 0)

            # Number of financial quarters (3-month blocks)
            divider = max(1, math.ceil(adjusted_months / 3))

            # If working-day exclusions apply, switch to day-based quarter calculation
            if exclude_weekdays or exclude_dates:
                total_working_days = count_days(
                    start_date,
                    end_date,
                    exclude_weekdays=exclude_weekdays,
                    exclude_dates=exclude_dates,
                )
                # Approx. 65 working days per quarter
                divider = max(1, math.ceil(total_working_days / 65))

        # CALENDAR QUARTER
        elif annotation_value_choice == ANNOTATION_VALUE_QUARTER:
            if exclude_weekdays or exclude_dates:
                total_working_days = count_days(
                    start_date,
                    end_date,
                    exclude_weekdays=exclude_weekdays,
                    exclude_dates=exclude_dates,
                )
                divider = max(1, math.ceil(total_working_days / 65))
            else:
                delta = relativedelta(end_date, start_date)
                divider = delta.years * 4 + math.ceil(delta.months / 3)
                if delta.days > 0 and delta.months % 3 == 0:
                    divider += 1
                divider = max(1, divider)
        # MONTH
        elif annotation_value_choice == ANNOTATION_VALUE_MONTH:
            rdate = relativedelta(end_date, start_date)
            divider = rdate.years * 12 + rdate.months
            if rdate.days > 0:
                divider += 1
            divider = divider or 1

        # WEEK
        elif annotation_value_choice == ANNOTATION_VALUE_WEEK:
            if exclude_weekdays or exclude_dates:
                total_working_days = count_days(
                    start_date,
                    end_date,
                    exclude_weekdays=exclude_weekdays,
                    exclude_dates=exclude_dates,
                )
                ex = {d for d in (exclude_weekdays or []) if 1 <= d <= 7}
                workdays_per_week = 7 - len(ex) if ex else 7
                if workdays_per_week == 0:
                    raise ReportError('All days are excluded from the week; cannot calculate number of weeks.')
                divider = math.ceil(total_working_days / workdays_per_week)
            else:
                monday1 = start_date - timedelta(days=start_date.weekday())
                monday2 = end_date - timedelta(days=end_date.weekday())
                divider = max(1, int((monday2 - monday1).days / 7))

        # DAY
        elif annotation_value_choice == ANNOTATION_VALUE_DAY:
            divider = count_days(
                start_date,
                end_date,
                exclude_weekdays=exclude_weekdays,
                exclude_dates=exclude_dates,
            )

        else:
            raise ReportError('unknown annotation_value_choice')

        return divider


class ChartBaseFieldForm(ReportBuilderFieldUtils, QueryBuilderForm):
    cancel_class = 'btn-secondary modal-cancel'

    def __init__(self, *args, **kwargs):
        self.django_field = None
        self.col_type_override = None
        super().__init__(*args, **kwargs)

    def cancel_button(self, css_class=cancel_class, **kwargs):
        commands = [{'function': 'close'}]
        return self.button('Cancel', commands, css_class, **kwargs)

    def get_report_type_details(self):
        data = json.loads(base64.b64decode(self.slug['data']))

        report_type = get_object_or_404(ReportType, pk=self.slug['report_type_id'])
        base_model = report_type.content_type.model_class()

        report_builder_class = get_report_builder_class(model=base_model, report_type=report_type)

        self.django_field, self.col_type_override, _, _ = self.get_field_details(
            base_model=base_model,
            field=data['field'],
            report_builder_class=report_builder_class,
        )

        return report_type, base_model

    def _get_query_builder_foreign_key_fields(
        self,
        base_model,
        report_builder_class,
        fields,
        prefix='',
        title_prefix='',
        previous_base_model=None,
    ):
        for include_field, include in report_builder_class.includes.items():
            app_label, model, report_builder_fields_str = include['model'].split('.')
            new_model = apps.get_model(app_label, model)
            if new_model != previous_base_model:
                new_report_builder_class = get_report_builder_class(
                    model=new_model, class_name=report_builder_fields_str
                )

                fields.append((prefix + include_field, title_prefix + include['title']))
                self._get_query_builder_foreign_key_fields(
                    base_model=new_model,
                    report_builder_class=new_report_builder_class,
                    fields=fields,
                    prefix=f'{prefix}{include_field}__',
                    title_prefix=f'{title_prefix}{include["title"]} --> ',
                    previous_base_model=base_model,
                )

    def setup_colour_field(self, form_fields, base_model, report_builder_class, name, data_attr, label=None):
        colour_fields = []
        self._get_colour_fields(
            base_model=base_model,
            report_builder_class=report_builder_class,
            fields=colour_fields,
        )
        dropdown_colour_fields = [['', '----']]
        for colour_field in colour_fields:
            dropdown_colour_fields.append([colour_field['id'], colour_field['text']])

        form_fields[name] = ChoiceField(
            choices=dropdown_colour_fields,
            required=False,
            widget=Select2(),
            label=label,
        )
        if name in data_attr:
            form_fields[name].initial = data_attr.get(name)



# Generated by Django 5.1.3 on 2025-11-21 14:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('advanced_report_builder', '0019_multicellstyle_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='multicellstyle',
            name='font_size',
            field=models.PositiveSmallIntegerField(blank=True, null=True),
        ),
    ]



# Generated by Django 5.1.3 on 2025-11-25 22:46

import django.db.models.deletion
import time_stamped_model.models
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('advanced_report_builder', '0020_alter_multicellstyle_font_size'),
    ]

    operations = [
        migrations.CreateModel(
            name='MultiValueHeldQuery',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', time_stamped_model.models.CreationDateTimeField(auto_now_add=True)),
                ('modified', time_stamped_model.models.ModificationDateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=200)),
                ('query', models.JSONField(blank=True, null=True)),
                ('multi_value_report', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='advanced_report_builder.multivaluereport')),
                ('report_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='advanced_report_builder.reporttype')),
            ],
            options={
                'get_latest_by': 'created',
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='multivaluereportcell',
            name='multi_value_held_query',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='advanced_report_builder.multivalueheldquery'),
        ),
    ]



from django.contrib import admin

from advanced_report_builder.models import (
    BarChartReport,
    CustomReport,
    Dashboard,
    DashboardReport,
    FunnelChartReport,
    KanbanReport,
    KanbanReportDescription,
    KanbanReportLane,
    LineChartReport,
    MultiCellStyle,
    MultiValueReport,
    MultiValueReportCell,
    MultiValueReportColumn,
    PieChartReport,
    Report,
    ReportQuery,
    ReportTag,
    ReportType,
    SingleValueReport,
    TableReport,
    Target,
)


@admin.register(ReportQuery)
class ReportQueryAdmin(admin.ModelAdmin):
    list_display = ('report', 'name', 'query')


class ReportQueryInline(admin.TabularInline):
    model = ReportQuery


@admin.register(Report)
class ReportAdmin(admin.ModelAdmin):
    list_display = ('name', 'instance_type', 'slug', 'slug_alias', 'version')
    exclude = ('slug',)

    inlines = [ReportQueryInline]


@admin.register(ReportType)
class ReportTypeAdmin(admin.ModelAdmin):
    list_display = ('name', 'content_type', 'report_builder_class_name')


@admin.register(TableReport)
class TableReportAdmin(admin.ModelAdmin):
    list_display = (
        'name',
        'slug',
    )
    exclude = (
        'instance_type',
        'slug',
    )
    inlines = [ReportQueryInline]


@admin.register(SingleValueReport)
class SingleValueReportAdmin(admin.ModelAdmin):
    list_display = (
        'name',
        'slug',
    )
    exclude = (
        'instance_type',
        'slug',
    )
    inlines = [ReportQueryInline]


@admin.register(BarChartReport)
class BarChartReportAdmin(admin.ModelAdmin):
    list_display = (
        'name',
        'slug',
    )
    exclude = (
        'instance_type',
        'slug',
    )
    inlines = [ReportQueryInline]


@admin.register(LineChartReport)
class LineChartReportAdmin(admin.ModelAdmin):
    list_display = (
        'name',
        'slug',
    )
    exclude = (
        'instance_type',
        'slug',
    )
    inlines = [ReportQueryInline]


@admin.register(PieChartReport)
class PieChartReportAdmin(admin.ModelAdmin):
    list_display = (
        'name',
        'slug',
    )
    exclude = (
        'instance_type',
        'slug',
    )
    inlines = [ReportQueryInline]


@admin.register(FunnelChartReport)
class FunnelChartReportAdmin(admin.ModelAdmin):
    list_display = (
        'name',
        'slug',
    )
    exclude = (
        'instance_type',
        'slug',
    )
    inlines = [ReportQueryInline]


@admin.register(CustomReport)
class CustomReportAdmin(admin.ModelAdmin):
    list_display = ('name', 'slug', 'output_type', 'view_name', 'settings')
    exclude = ('instance_type', 'slug', 'report_type')


class KanbanReportDescriptionInline(admin.TabularInline):
    model = KanbanReportDescription


class KanbanReportLaneInline(admin.TabularInline):
    model = KanbanReportLane


@admin.register(KanbanReport)
class KanbanReportAdmin(admin.ModelAdmin):
    list_display = (
        'name',
        'slug',
    )
    exclude = ('instance_type', 'slug', 'report_type')
    inlines = [KanbanReportLaneInline, KanbanReportDescriptionInline]


class DashboardReportInline(admin.TabularInline):
    model = DashboardReport


@admin.register(Dashboard)
class DashboardAdmin(admin.ModelAdmin):
    list_display = (
        'slug',
        'name',
    )

    exclude = [
        'slug',
    ]

    inlines = [DashboardReportInline]


@admin.register(ReportTag)
class ReportTagAdmin(admin.ModelAdmin):
    list_display = ('name', 'order')


@admin.register(Target)
class TargetAdmin(admin.ModelAdmin):
    list_display = ('pk', 'name', 'target_type', 'default_value')


class MultiCellStyleInline(admin.TabularInline):
    model = MultiCellStyle
    extra = 0


class MultiValueReportColumnInline(admin.TabularInline):
    model = MultiValueReportColumn
    extra = 0
    fields = ('column', 'width_type', 'width')
    ordering = ('column',)


class MultiValueReportCellInline(admin.TabularInline):
    model = MultiValueReportCell
    extra = 0
    fields = (
        'row',
        'column',
        'col_span',
        'row_span',
        'multi_value_type',
        'text',
        'multi_cell_style',
        'report_type',
        'field',
        'numerator',
        'prefix',
        'decimal_places',
        'show_breakdown',
        'breakdown_fields',
        'average_scale',
        'average_start_period',
        'average_end_period',
        'label',
        'query_data',
        'extra_query_data',
    )
    ordering = ('row', 'column')


@admin.register(MultiValueReport)
class MultiValueReportAdmin(admin.ModelAdmin):
    list_display = ('id', 'rows', 'columns', 'default_multi_cell_style')
    inlines = [
        MultiCellStyleInline,
        MultiValueReportColumnInline,
        MultiValueReportCellInline,
    ]


@admin.register(MultiCellStyle)
class MultiCellStyleAdmin(admin.ModelAdmin):
    list_display = ('name', 'multi_value_report', 'align_type', 'bold', 'italic', 'font_size')
    list_filter = ('multi_value_report', 'align_type', 'bold', 'italic')
    search_fields = ('name',)


@admin.register(MultiValueReportColumn)
class MultiValueReportColumnAdmin(admin.ModelAdmin):
    list_display = ('multi_value_report', 'column', 'width_type', 'width')
    list_filter = ('multi_value_report', 'width_type')
    ordering = ('multi_value_report', 'column')


@admin.register(MultiValueReportCell)
class MultiValueReportCellAdmin(admin.ModelAdmin):
    list_display = ('multi_value_report', 'row', 'column', 'multi_value_type', 'multi_cell_style')
    list_filter = ('multi_value_report', 'multi_value_type')
    search_fields = ('text', 'label')



# Generated by Django 5.1.3 on 2025-07-04 14:30

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('advanced_report_builder', '0017_alter_barchartreport_options_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='reportquery',
            name='target',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='advanced_report_builder.target'),
        ),
        migrations.AddField(
            model_name='target',
            name='period_type',
            field=models.PositiveSmallIntegerField(choices=[(1, 'Daily'), (2, 'Weekly'), (3, 'Monthly'), (4, 'Quarterly')], default=3),
        ),
    ]



from dataclasses import dataclass
from enum import Enum

from django.apps import apps
from django.contrib.auth.models import AbstractUser
from django.db import models
from django_datatables.columns import ManyToManyColumn

from advanced_report_builder.column_types import DATE_FIELDS
from advanced_report_builder.columns import FilterForeignKeyColumn
from advanced_report_builder.field_utils import ReportBuilderFieldUtils
from advanced_report_builder.globals import FieldType
from advanced_report_builder.utils import get_report_builder_class
from advanced_report_builder.variable_date import VariableDate


@dataclass(slots=True)
class FieldDetail:
    field_type: FieldType
    field: str
    title: str
    django_field: models.Field | None = None
    column: str | None = None
    prefix: str | None = None
    full_field_name: str | None = None
    column_id: int | None = None

    def to_dict(self):
        return self.__dict__


class FieldTypes(ReportBuilderFieldUtils):
    class OperatorFieldType(Enum):
        STRING = 1
        NUMBER = 2
        DATE = 3
        BOOLEAN = 4
        MULTIPLE_CHOICE = 5
        FOREIGN_KEY = 6
        ABSTRACT_USER = 7
        PART_DATE = 8

        COMPARE_STRING = 9
        COMPARE_NUMBER = 10
        COMPARE_DATE = 11
        COMPARE_BOOLEAN = 12
        WEEKDAYS = 13

    def get_operator(self, field_type):
        operators = {
            self.OperatorFieldType.STRING: [
                'equal',
                'not_equal',
                'contains',
                'not_contains',
                'begins_with',
                'not_begins_with',
                'ends_with',
                'not_ends_with',
            ],
            self.OperatorFieldType.NUMBER: [
                'equal',
                'not_equal',
                'less',
                'less_or_equal',
                'greater',
                'greater_or_equal',
            ],
            self.OperatorFieldType.DATE: [
                'equal',
                'not_equal',
                'less',
                'less_or_equal',
                'greater',
                'greater_or_equal',
                'is_null',
                'is_not_null',
            ],
            self.OperatorFieldType.PART_DATE: ['equal', 'not_equal', 'is_null', 'is_not_null'],
            self.OperatorFieldType.BOOLEAN: ['equal', 'not_equal'],
            self.OperatorFieldType.MULTIPLE_CHOICE: ['in', 'not_in'],
            self.OperatorFieldType.FOREIGN_KEY: ['is_null', 'is_not_null'],
            self.OperatorFieldType.ABSTRACT_USER: ['equal', 'not_equal'],
            self.OperatorFieldType.WEEKDAYS: ['equal', 'not_equal'],
            self.OperatorFieldType.COMPARE_STRING: ['equal', 'not_equal'],
            self.OperatorFieldType.COMPARE_NUMBER: [
                'equal',
                'not_equal',
                'less',
                'less_or_equal',
                'greater',
                'greater_or_equal',
            ],
            self.OperatorFieldType.COMPARE_DATE: [
                'equal',
                'not_equal',
                'less',
                'less_or_equal',
                'greater',
                'greater_or_equal',
            ],
            self.OperatorFieldType.COMPARE_BOOLEAN: ['equal', 'not_equal'],
        }
        return operators.get(field_type)

    def get_filters(self, fields, query_builder_filters, field_results_types):
        for field_detail in fields:
            field_type = field_detail.field_type

            if field_type == FieldType.NULL_FIELD:
                self.get_foreign_key_null_field(
                    query_builder_filters=query_builder_filters,
                    field=field_detail.field,
                    title=field_detail.title,
                )

            elif field_type == FieldType.ABSTRACT_USER:
                self.get_abstract_user_field(
                    query_builder_filters=query_builder_filters,
                    field=field_detail.field,
                    title=field_detail.title,
                )

            elif field_type == FieldType.FILTER_FOREIGN_KEY:
                query_builder_filters.append(
                    {
                        'id': field_detail.column_id,
                        'label': field_detail.title,
                        'field': field_detail.full_field_name,
                        'type': 'string',
                        'input': 'select',
                        'multiple': True,
                        'values': field_detail.column.get_query_options(),
                        'operators': self.get_operator(self.OperatorFieldType.MULTIPLE_CHOICE),
                    }
                )
            elif field_type == FieldType.STRING:
                query_builder_filters.append(
                    {
                        'id': field_detail.column_id,
                        'label': f'{field_detail.title} (Field vs Value)',
                        'field': field_detail.full_field_name,
                        'type': 'string',
                        'operators': self.get_operator(self.OperatorFieldType.STRING),
                        'validation': {'allow_empty_value': True},
                    }
                )
                query_builder_filters.append(
                    {
                        'id': f'{field_detail.column_id}__field_vs_field',
                        'label': f'{field_detail.title} (Field vs Field)',
                        'field': field_detail.full_field_name,
                        'input': 'select',
                        'operators': self.get_operator(self.OperatorFieldType.COMPARE_STRING),
                        'values': field_results_types[FieldType.STRING],
                    }
                )
            elif field_type == FieldType.NUMBER:
                if field_detail.django_field.choices is None:
                    query_builder_filter = {
                        'id': field_detail.column_id,
                        'label': f'{field_detail.title} (Field vs Value)',
                        'field': field_detail.full_field_name,
                        'type': 'integer',
                        'operators': self.get_operator(self.OperatorFieldType.NUMBER),
                    }
                    query_builder_filters.append(
                        {
                            'id': f'{field_detail.column_id}__field_vs_field',
                            'label': f'{field_detail.title} (Field vs Field)',
                            'field': field_detail.full_field_name,
                            'input': 'select',
                            'operators': self.get_operator(self.OperatorFieldType.COMPARE_NUMBER),
                            'values': field_results_types[FieldType.NUMBER],
                        }
                    )
                else:
                    query_builder_filter = {
                        'id': field_detail.column_id,
                        'label': field_detail.title,
                        'field': field_detail.full_field_name,
                        'type': 'integer',
                        'input': 'select',
                        'multiple': True,
                        'values': dict(field_detail.django_field.choices),
                        'operators': self.get_operator(self.OperatorFieldType.MULTIPLE_CHOICE),
                    }
                query_builder_filters.append(query_builder_filter)
            elif field_type == FieldType.BOOLEAN:
                query_builder_filters.append(
                    {
                        'id': field_detail.column_id,
                        'label': f'{field_detail.title} (Field vs Value)',
                        'field': field_detail.full_field_name,
                        'input': 'select',
                        'operators': self.get_operator(self.OperatorFieldType.BOOLEAN),
                        'values': {'0': 'False', '1': 'True'},
                    }
                )
                query_builder_filters.append(
                    {
                        'id': f'{field_detail.column_id}__field_vs_field',
                        'label': f'{field_detail.title} (Field vs Field)',
                        'field': field_detail.full_field_name,
                        'input': 'select',
                        'operators': self.get_operator(self.OperatorFieldType.COMPARE_BOOLEAN),
                        'values': field_results_types[FieldType.BOOLEAN],
                    }
                )

            elif field_type == FieldType.DATE:
                self.get_date_field(
                    field_results_types=field_results_types,
                    column_id=field_detail.column_id,
                    query_builder_filters=query_builder_filters,
                    field=field_detail.full_field_name,
                    title=field_detail.title,
                )
            elif field_type == FieldType.MANY_TO_MANY:
                choices = dict(field_detail.column.options['lookup'])
                query_builder_filter = {
                    'id': field_detail.column_id,
                    'label': field_detail.title,
                    'field': field_detail.prefix + field_detail.column.field_id,
                    'type': 'integer',
                    'input': 'select',
                    'multiple': True,
                    'values': choices,
                    'operators': self.get_operator(self.OperatorFieldType.MULTIPLE_CHOICE),
                }
                query_builder_filters.append(query_builder_filter)

    def get_foreign_key_null_field(self, query_builder_filters, field, title):
        query_builder_filters.append(
            {
                'id': field,
                'label': title,
                'field': field,
                'type': 'integer',
                'operators': self.get_operator(self.OperatorFieldType.FOREIGN_KEY),
                'validation': {'allow_empty_value': True},
            }
        )

    def get_date_field(self, field_results_types, column_id, query_builder_filters, field, title):
        variable_date = VariableDate()
        query_builder_filter = {
            'id': f'{column_id}__variable_date',
            'label': f'{title} (Variable)',
            'field': field,
            'operators': self.get_operator(self.OperatorFieldType.DATE),
            'input': 'select',
            'values': variable_date.get_variable_date_filter_values(),
        }
        query_builder_filters.append(query_builder_filter)
        query_builder_filter = {
            'id': f'{column_id}__variable_year',
            'label': f'{title} (Year)',
            'field': field,
            'operators': self.get_operator(self.OperatorFieldType.DATE),
            'input': 'select',
            'values': variable_date.get_date_filter_years(),
        }
        query_builder_filters.append(query_builder_filter)
        query_builder_filter = {
            'id': f'{column_id}__variable_month',
            'label': f'{title} (Month)',
            'field': field,
            'operators': self.get_operator(self.OperatorFieldType.DATE),
            'input': 'select',
            'values': variable_date.get_date_filter_months(),
        }
        query_builder_filters.append(query_builder_filter)

        query_builder_filter = {
            'id': f'{column_id}__variable_quarter',
            'label': f'{title} (Quarter)',
            'field': field,
            'operators': self.get_operator(self.OperatorFieldType.PART_DATE),
            'input': 'select',
            'values': variable_date.get_date_filter_quarters(),
        }
        query_builder_filters.append(query_builder_filter)

        query_builder_filter = {
            'id': f'{column_id}__variable_day',
            'label': f'{title} (Day)',
            'field': field,
            'operators': self.get_operator(self.OperatorFieldType.WEEKDAYS),
            'input': 'select',
            'values': {
                1: 'Sunday',
                2: 'Monday',
                3: 'Tuesday',
                4: 'Wednesday',
                5: 'Thursday',
                6: 'Friday',
                7: 'Saturday',
            },
        }
        query_builder_filters.append(query_builder_filter)

        query_builder_filters.append(
            {
                'id': f'{column_id}__field_vs_field',
                'label': f'{title} (Field vs Field)',
                'field': field,
                'input': 'select',
                'operators': self.get_operator(self.OperatorFieldType.COMPARE_DATE),
                'values': field_results_types[FieldType.DATE],
            }
        )

    def get_abstract_user_field(self, query_builder_filters, field, title):
        query_builder_filters.append(
            {
                'id': f'{field}__logged_in_user',
                'label': f'{title} (Logged in user)',
                'field': field,
                'input': 'select',
                'operators': self.get_operator(self.OperatorFieldType.ABSTRACT_USER),
                'values': {'0': 'False', '1': 'True'},
            }
        )

    def get_field_types(
        self,
        field_results,
        field_results_types,
        base_model,
        report_builder_class,
        prefix='',
        title_prefix='',
        previous_base_model=None,
        show_includes=True,
    ):
        for report_builder_field in report_builder_class.fields:
            if (
                not isinstance(report_builder_field, str)
                or report_builder_field not in report_builder_class.exclude_search_fields
            ):
                django_field, _, columns, _ = self.get_field_details(
                    base_model=base_model,
                    field=report_builder_field,
                    report_builder_class=report_builder_class,
                )

                for column in columns:
                    field = prefix + column.column_name
                    selected_field_type = None
                    if django_field is not None and not isinstance(column, ManyToManyColumn):
                        if isinstance(column, FilterForeignKeyColumn):
                            selected_field_type = FieldType.FILTER_FOREIGN_KEY
                        elif isinstance(django_field, models.CharField | models.TextField | models.EmailField):
                            selected_field_type = FieldType.STRING
                        elif isinstance(
                            django_field,
                            models.IntegerField
                            | models.PositiveSmallIntegerField
                            | models.PositiveIntegerField
                            | models.FloatField,
                        ):
                            selected_field_type = FieldType.NUMBER
                        elif isinstance(django_field, models.BooleanField):
                            selected_field_type = FieldType.BOOLEAN
                        elif isinstance(django_field, DATE_FIELDS):
                            selected_field_type = FieldType.DATE
                    elif isinstance(column, ManyToManyColumn):
                        selected_field_type = FieldType.MANY_TO_MANY

                    if isinstance(column.field, list):
                        full_field_name = prefix + column.field[0]
                    elif column.field is not None:
                        full_field_name = prefix + column.field
                    else:
                        full_field_name = prefix + field
                    column_id = prefix + field

                    if selected_field_type is not None:
                        title = title_prefix + column.title
                        field_results_types[selected_field_type][column_id] = title
                        field_results.append(
                            FieldDetail(
                                django_field=django_field,
                                column=column,
                                title=title,
                                field=column.column_name,
                                prefix=prefix,
                                field_type=selected_field_type,
                                full_field_name=full_field_name,
                                column_id=column_id,
                            )
                        )

        if show_includes:
            for include_field, include in report_builder_class.includes.items():
                app_label, model, report_builder_fields_str = include['model'].split('.')
                new_model = apps.get_model(app_label, model)
                new_report_builder_class = get_report_builder_class(
                    model=new_model, class_name=report_builder_fields_str
                )

                if new_model != previous_base_model:
                    _foreign_key = getattr(base_model, include_field, None)

                    add_null_field = _foreign_key.field.null if hasattr(_foreign_key, 'field') else True

                    if add_null_field:
                        field = prefix + include_field
                        title = title_prefix + include['title']
                        field_results_types[FieldType.NULL_FIELD][field] = title
                        field_results.append(FieldDetail(field=field, title=title, field_type=FieldType.NULL_FIELD))
                    if issubclass(new_model, AbstractUser):
                        field = prefix + include_field
                        title = title_prefix + include['title']
                        field_results_types[FieldType.NULL_FIELD][field] = title
                        field_results.append(FieldDetail(field=field, title=title, field_type=FieldType.ABSTRACT_USER))
                    self.get_field_types(
                        field_results=field_results,
                        field_results_types=field_results_types,
                        base_model=new_model,
                        report_builder_class=new_report_builder_class,
                        prefix=f'{prefix}{include_field}__',
                        title_prefix=f'{include["title"]} --> ',
                        previous_base_model=base_model,
                        show_includes=include.get('show_includes', True),
                    )



# Generated by Django 5.1.3 on 2025-11-20 12:01

import django.db.models.deletion
import django_modals.model_fields.colour
import time_stamped_model.models
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('advanced_report_builder', '0018_reportquery_target_target_period_type'),
    ]

    operations = [
        migrations.CreateModel(
            name='MultiCellStyle',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', time_stamped_model.models.CreationDateTimeField(auto_now_add=True)),
                ('modified', time_stamped_model.models.ModificationDateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=200)),
                ('align_type', models.IntegerField(choices=[(0, 'Left'), (1, 'Centre'), (2, 'Right')], default=0)),
                ('bold', models.BooleanField(default=False)),
                ('italic', models.BooleanField(default=False)),
                ('font_size', models.PositiveSmallIntegerField()),
                ('font_colour', django_modals.model_fields.colour.ColourField(blank=True, max_length=10, null=True)),
                ('background_colour', django_modals.model_fields.colour.ColourField(blank=True, max_length=10, null=True)),
            ],
            options={
                'get_latest_by': 'created',
                'abstract': False,
            },
        ),
        migrations.AlterField(
            model_name='singlevaluereport',
            name='single_value_type',
            field=models.PositiveSmallIntegerField(choices=[(1, 'Count'), (2, 'Sum'), (3, 'Count & Sum'), (4, 'Percent'), (5, 'Percent from Count'), (6, 'Average Sum from Count'), (7, 'Average Sum over Time'), (8, 'Average Sum over Time (Excluding Weekends)')], default=1),
        ),
        migrations.CreateModel(
            name='MultiValueReport',
            fields=[
                ('report_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='advanced_report_builder.report')),
                ('rows', models.PositiveSmallIntegerField()),
                ('columns', models.PositiveSmallIntegerField()),
                ('default_multi_cell_style', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='advanced_report_builder.multicellstyle')),
            ],
            options={
                'get_latest_by': 'created',
                'abstract': False,
            },
            bases=('advanced_report_builder.report',),
        ),
        migrations.AddField(
            model_name='multicellstyle',
            name='multi_value_report',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='advanced_report_builder.multivaluereport'),
        ),
        migrations.CreateModel(
            name='MultiValueReportColumn',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', time_stamped_model.models.CreationDateTimeField(auto_now_add=True)),
                ('modified', time_stamped_model.models.ModificationDateTimeField(auto_now=True)),
                ('column', models.PositiveSmallIntegerField()),
                ('width_type', models.PositiveSmallIntegerField(choices=[(0, '%'), (1, 'Pixels')], default=0)),
                ('width', models.PositiveSmallIntegerField()),
                ('multi_value_report', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='advanced_report_builder.multivaluereport')),
            ],
            options={
                'get_latest_by': 'created',
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='MultiValueReportCell',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', time_stamped_model.models.CreationDateTimeField(auto_now_add=True)),
                ('modified', time_stamped_model.models.ModificationDateTimeField(auto_now=True)),
                ('row', models.PositiveSmallIntegerField()),
                ('column', models.PositiveSmallIntegerField()),
                ('col_span', models.PositiveSmallIntegerField(default=1)),
                ('row_span', models.PositiveSmallIntegerField(default=1)),
                ('multi_value_type', models.IntegerField(choices=[(0, 'Static Text'), (1, 'Count'), (2, 'Sum'), (4, 'Percent'), (5, 'Percent from Count'), (6, 'Average Sum from Count'), (7, 'Average Sum over Time'), (8, 'Average Sum over Time (Excluding Weekends)'), (9, 'Equation')], default=0)),
                ('text', models.TextField(blank=True, null=True)),
                ('field', models.CharField(blank=True, max_length=200, null=True)),
                ('numerator', models.CharField(blank=True, max_length=200, null=True)),
                ('prefix', models.CharField(blank=True, max_length=64, null=True)),
                ('decimal_places', models.IntegerField(default=0)),
                ('show_breakdown', models.BooleanField(default=False)),
                ('breakdown_fields', models.JSONField(blank=True, null=True)),
                ('average_scale', models.PositiveSmallIntegerField(blank=True, choices=[(1, 'Year'), (2, 'Quarter'), (3, 'Month'), (4, 'Week'), (5, 'Day')], null=True)),
                ('average_start_period', models.PositiveSmallIntegerField(blank=True, null=True)),
                ('average_end_period', models.PositiveSmallIntegerField(blank=True, null=True)),
                ('label', models.CharField(blank=True, max_length=256, null=True)),
                ('query_data', models.JSONField(blank=True, null=True)),
                ('extra_query_data', models.JSONField(blank=True, null=True)),
                ('multi_cell_style', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='advanced_report_builder.multicellstyle')),
                ('multi_value_report', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='advanced_report_builder.multivaluereport')),
                ('report_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='advanced_report_builder.reporttype')),
            ],
            options={
                'constraints': [models.UniqueConstraint(fields=('multi_value_report', 'row', 'column'), name='multi_value_report_cell_unique')],
            },
        ),
    ]



from django_menus.menu import MenuItem

from advanced_report_builder.views.custom import CustomBaseView


class ErrorPodView(CustomBaseView):
    template_name = 'advanced_report_builder/error.html'
    report_type_slug = 'company'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.enable_queries = False

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['error_message'] = self.extra_kwargs.get('error_message')
        return context

    def pod_report_menu(self):
        report_type = self.get_report_type()
        return [
            ('report_builder_examples:index', 'Back', {'css_classes': 'btn-secondary'}),
            MenuItem(
                f'advanced_report_builder:custom_modal,pk-{self.report.id}-report_type-{report_type.id}',
                menu_display='Edit',
                font_awesome='fas fa-pencil-alt',
            ),
        ]

    def setup_menu(self):
        super().setup_menu()
        if not self.show_toolbar:
            return

        if self.dashboard_report and self.enable_edit:
            report_menu = self.pod_dashboard_edit_menu()
        elif self.dashboard_report and not self.enable_edit:
            report_menu = self.pod_dashboard_view_menu()
        else:
            view = self.extra_kwargs.get('report_cls')(*self.args, **self.kwargs)
            report_menu = view.edit_report_menu(request=self.request, chart_report_id=self.report.id, slug_str='')
        if self.enable_queries:
            self.add_menu('button_menu', 'button_group').add_items(
                *report_menu,
                *self.queries_menu(report=self.report, dashboard_report=self.dashboard_report),
            )
        else:
            self.add_menu('button_menu', 'button_group').add_items(
                *report_menu,
            )



import calendar
from calendar import monthrange
from datetime import date, datetime, timedelta

from date_offset.date_offset import DateOffset
from dateutil.relativedelta import relativedelta


class VariableDate:
    RANGE_TYPE_TODAY = 1
    RANGE_TYPE_YESTERDAY = 2
    RANGE_TYPE_TOMORROW = 3
    RANGE_TYPE_THIS_WEEK = 4
    RANGE_TYPE_NEXT_WEEK = 5
    RANGE_TYPE_2_WEEKS_TIME = 6
    RANGE_TYPE_3_WEEKS_TIME = 34
    RANGE_TYPE_4_WEEKS_TIME = 35
    RANGE_TYPE_5_WEEKS_TIME = 36
    RANGE_TYPE_6_WEEKS_TIME = 37
    RANGE_TYPE_THIS_MONTH = 7
    RANGE_TYPE_NEXT_MONTH = 8
    RANGE_TYPE_LAST_MONTH = 9
    RANGE_TYPE_NEXT_7_DAYS = 10
    RANGE_TYPE_LAST_7_DAYS = 11
    RANGE_TYPE_NEXT_14_DAYS = 12
    RANGE_TYPE_LAST_14_DAYS = 13
    RANGE_TYPE_NEXT_28_DAYS = 14
    RANGE_TYPE_LAST_28_DAYS = 15
    RANGE_TYPE_THIS_YEAR = 16
    RANGE_TYPE_LAST_YEAR = 17
    RANGE_TYPE_LAST_12_MONTHS = 38
    RANGE_TYPE_NEXT_12_MONTHS = 39
    RANGE_TYPE_LAST_WEEK = 18

    RANGE_TYPE_NEXT_2_MONTHS = 24
    RANGE_TYPE_LAST_2_MONTHS = 25
    RANGE_TYPE_NEXT_3_MONTHS = 26
    RANGE_TYPE_LAST_3_MONTHS = 27
    RANGE_TYPE_NEXT_6_MONTHS = 28
    RANGE_TYPE_LAST_6_MONTHS = 29
    RANGE_TYPE_NEXT_60_DAYS = 30
    RANGE_TYPE_LAST_60_DAYS = 31
    RANGE_TYPE_NEXT_90_DAYS = 32
    RANGE_TYPE_LAST_90_DAYS = 33
    RANGE_TYPE_LAST_18_MONTHS = 23
    RANGE_TYPE_LAST_24_MONTHS = 40
    RANGE_TYPE_NEXT_24_MONTHS = 41
    RANGE_TYPE_LAST_36_MONTHS = 42
    RANGE_TYPE_NEXT_36_MONTHS = 43

    RANGE_TYPE_2_WEEKS_AGO = 44
    RANGE_TYPE_3_WEEKS_AGO = 45
    RANGE_TYPE_4_WEEKS_AGO = 46
    RANGE_TYPE_5_WEEKS_AGO = 47
    RANGE_TYPE_6_WEEKS_AGO = 48

    RANGE_TYPE_THIS_YEAR = 16
    RANGE_TYPE_LAST_YEAR = 17

    # Financial year ranges
    RANGE_TYPE_LAST_FINANCIAL_YEAR = 49
    RANGE_TYPE_THIS_FINANCIAL_YEAR = 50
    RANGE_TYPE_NEXT_FINANCIAL_YEAR = 51

    # Last FY quarters
    RANGE_TYPE_LAST_FINANCIAL_YEAR_Q1 = 52
    RANGE_TYPE_LAST_FINANCIAL_YEAR_Q2 = 53
    RANGE_TYPE_LAST_FINANCIAL_YEAR_Q3 = 54
    RANGE_TYPE_LAST_FINANCIAL_YEAR_Q4 = 55

    # This FY quarters
    RANGE_TYPE_THIS_FINANCIAL_YEAR_Q1 = 56
    RANGE_TYPE_THIS_FINANCIAL_YEAR_Q2 = 57
    RANGE_TYPE_THIS_FINANCIAL_YEAR_Q3 = 58
    RANGE_TYPE_THIS_FINANCIAL_YEAR_Q4 = 59

    # Next FY quarters
    RANGE_TYPE_NEXT_FINANCIAL_YEAR_Q1 = 60
    RANGE_TYPE_NEXT_FINANCIAL_YEAR_Q2 = 61
    RANGE_TYPE_NEXT_FINANCIAL_YEAR_Q3 = 62
    RANGE_TYPE_NEXT_FINANCIAL_YEAR_Q4 = 63

    # Current FY quarters
    RANGE_TYPE_LAST_FINANCIAL_QUARTER = 64
    RANGE_TYPE_CURRENT_FINANCIAL_QUARTER = 65
    RANGE_TYPE_NEXT_FINANCIAL_QUARTER = 66
    RANGE_TYPE_LAST_YEAR_SAME_FINANCIAL_QUARTER = 67

    # This calendar quarter
    RANGE_TYPE_LAST_CALENDAR_QUARTER = 68
    RANGE_TYPE_THIS_CALENDAR_QUARTER = 69
    RANGE_TYPE_NEXT_CALENDAR_QUARTER = 70

    # Last calendar quarters
    RANGE_TYPE_LAST_CALENDAR_YEAR_Q1 = 71
    RANGE_TYPE_LAST_CALENDAR_YEAR_Q2 = 72
    RANGE_TYPE_LAST_CALENDAR_YEAR_Q3 = 73
    RANGE_TYPE_LAST_CALENDAR_YEAR_Q4 = 74

    # This calendar quarters
    RANGE_TYPE_THIS_CALENDAR_YEAR_Q1 = 75
    RANGE_TYPE_THIS_CALENDAR_YEAR_Q2 = 76
    RANGE_TYPE_THIS_CALENDAR_YEAR_Q3 = 77
    RANGE_TYPE_THIS_CALENDAR_YEAR_Q4 = 78

    # Next calendar quarters
    RANGE_TYPE_NEXT_CALENDAR_YEAR_Q1 = 79
    RANGE_TYPE_NEXT_CALENDAR_YEAR_Q2 = 80
    RANGE_TYPE_NEXT_CALENDAR_YEAR_Q3 = 81
    RANGE_TYPE_NEXT_CALENDAR_YEAR_Q4 = 82

    RANGE_TYPE_CHOICES = (
        (RANGE_TYPE_LAST_36_MONTHS, 'Last 36 months'),
        (RANGE_TYPE_LAST_24_MONTHS, 'Last 24 months'),
        (RANGE_TYPE_LAST_18_MONTHS, 'Last 18 months'),
        (RANGE_TYPE_LAST_12_MONTHS, 'Last 12 months'),
        (RANGE_TYPE_LAST_6_MONTHS, 'Last 6 months'),
        (RANGE_TYPE_LAST_3_MONTHS, 'Last 3 months'),
        (RANGE_TYPE_LAST_90_DAYS, 'Last 90 days'),
        (RANGE_TYPE_LAST_2_MONTHS, 'Last 2 months'),
        (RANGE_TYPE_LAST_60_DAYS, 'Last 60 days'),
        (RANGE_TYPE_LAST_MONTH, 'Last month'),
        (RANGE_TYPE_LAST_28_DAYS, 'Last 28 days'),
        (RANGE_TYPE_LAST_14_DAYS, 'Last 14 days'),
        (RANGE_TYPE_LAST_7_DAYS, 'Last 7 days'),
        (RANGE_TYPE_6_WEEKS_AGO, '6 weeks ago'),
        (RANGE_TYPE_5_WEEKS_AGO, '5 weeks ago'),
        (RANGE_TYPE_4_WEEKS_AGO, '4 weeks ago'),
        (RANGE_TYPE_3_WEEKS_AGO, '3 weeks ago'),
        (RANGE_TYPE_2_WEEKS_AGO, '2 weeks ago'),
        (RANGE_TYPE_LAST_WEEK, 'Last week'),
        (RANGE_TYPE_YESTERDAY, 'Yesterday'),
        (RANGE_TYPE_TODAY, 'Today'),
        (RANGE_TYPE_TOMORROW, 'Tomorrow'),
        (RANGE_TYPE_THIS_WEEK, 'This week'),
        (RANGE_TYPE_NEXT_7_DAYS, 'Next 7 days'),
        (RANGE_TYPE_NEXT_14_DAYS, 'Next 14 days'),
        (RANGE_TYPE_NEXT_WEEK, 'Next week'),
        (RANGE_TYPE_2_WEEKS_TIME, '2 weeks time'),
        (RANGE_TYPE_3_WEEKS_TIME, '3 weeks time'),
        (RANGE_TYPE_4_WEEKS_TIME, '4 weeks time'),
        (RANGE_TYPE_5_WEEKS_TIME, '5 weeks time'),
        (RANGE_TYPE_6_WEEKS_TIME, '6 weeks time'),
        (RANGE_TYPE_THIS_MONTH, 'This month'),
        (RANGE_TYPE_NEXT_28_DAYS, 'Next 28 days'),
        (RANGE_TYPE_NEXT_MONTH, 'Next month'),
        (RANGE_TYPE_NEXT_2_MONTHS, 'Next 2 months'),
        (RANGE_TYPE_NEXT_60_DAYS, 'Next 60 days'),
        (RANGE_TYPE_NEXT_3_MONTHS, 'Next 3 months'),
        (RANGE_TYPE_NEXT_90_DAYS, 'Next 90 days'),
        (RANGE_TYPE_NEXT_6_MONTHS, 'Next 6 months'),
        (RANGE_TYPE_NEXT_12_MONTHS, 'Next 12 months'),
        (RANGE_TYPE_NEXT_24_MONTHS, 'Next 24 months'),
        (RANGE_TYPE_NEXT_36_MONTHS, 'Next 36 months'),
        # Financial year full ranges
        (RANGE_TYPE_LAST_FINANCIAL_YEAR, 'Last financial year'),
        (RANGE_TYPE_THIS_FINANCIAL_YEAR, 'This financial year'),
        (RANGE_TYPE_NEXT_FINANCIAL_YEAR, 'Next financial year'),
        # Last FY quarters
        (RANGE_TYPE_LAST_FINANCIAL_YEAR_Q1, 'Last financial year Q1'),
        (RANGE_TYPE_LAST_FINANCIAL_YEAR_Q2, 'Last financial year Q2'),
        (RANGE_TYPE_LAST_FINANCIAL_YEAR_Q3, 'Last financial year Q3'),
        (RANGE_TYPE_LAST_FINANCIAL_YEAR_Q4, 'Last financial year Q4'),
        # This FY quarters
        (RANGE_TYPE_THIS_FINANCIAL_YEAR_Q1, 'This financial year Q1'),
        (RANGE_TYPE_THIS_FINANCIAL_YEAR_Q2, 'This financial year Q2'),
        (RANGE_TYPE_THIS_FINANCIAL_YEAR_Q3, 'This financial year Q3'),
        (RANGE_TYPE_THIS_FINANCIAL_YEAR_Q4, 'This financial year Q4'),
        # Next FY quarters
        (RANGE_TYPE_NEXT_FINANCIAL_YEAR_Q1, 'Next financial year Q1'),
        (RANGE_TYPE_NEXT_FINANCIAL_YEAR_Q2, 'Next financial year Q2'),
        (RANGE_TYPE_NEXT_FINANCIAL_YEAR_Q3, 'Next financial year Q3'),
        (RANGE_TYPE_NEXT_FINANCIAL_YEAR_Q4, 'Next financial year Q4'),
        # Current FY quarters
        (RANGE_TYPE_LAST_FINANCIAL_QUARTER, 'Last financial quarter'),
        (RANGE_TYPE_CURRENT_FINANCIAL_QUARTER, 'Current financial quarter'),
        (RANGE_TYPE_NEXT_FINANCIAL_QUARTER, 'Next financial quarter'),
        (RANGE_TYPE_LAST_YEAR_SAME_FINANCIAL_QUARTER, 'Same quarter last financial year'),
        # Calendar year full ranges
        (RANGE_TYPE_LAST_CALENDAR_QUARTER, 'Last calendar quarter'),
        (RANGE_TYPE_THIS_CALENDAR_QUARTER, 'This calendar quarter'),
        (RANGE_TYPE_NEXT_CALENDAR_QUARTER, 'Next calendar quarter'),
        # Last calendar year quarters
        (RANGE_TYPE_LAST_CALENDAR_YEAR_Q1, 'Last calendar year Q1'),
        (RANGE_TYPE_LAST_CALENDAR_YEAR_Q2, 'Last calendar year Q2'),
        (RANGE_TYPE_LAST_CALENDAR_YEAR_Q3, 'Last calendar year Q3'),
        (RANGE_TYPE_LAST_CALENDAR_YEAR_Q4, 'Last calendar year Q4'),
        # This calendar year quarters
        (RANGE_TYPE_THIS_CALENDAR_YEAR_Q1, 'This calendar year Q1'),
        (RANGE_TYPE_THIS_CALENDAR_YEAR_Q2, 'This calendar year Q2'),
        (RANGE_TYPE_THIS_CALENDAR_YEAR_Q3, 'This calendar year Q3'),
        (RANGE_TYPE_THIS_CALENDAR_YEAR_Q4, 'This calendar year Q4'),
        # Next calendar year quarters
        (RANGE_TYPE_NEXT_CALENDAR_YEAR_Q1, 'Next calendar year Q1'),
        (RANGE_TYPE_NEXT_CALENDAR_YEAR_Q2, 'Next calendar year Q2'),
        (RANGE_TYPE_NEXT_CALENDAR_YEAR_Q3, 'Next calendar year Q3'),
        (RANGE_TYPE_NEXT_CALENDAR_YEAR_Q4, 'Next calendar year Q4'),
    )

    def get_variable_dates(self, range_type, financial_year_start_month=1):  # financial_year_start_month 1=jan
        today = date.today()
        start_of_this_week = today - timedelta(days=today.weekday())
        if range_type == self.RANGE_TYPE_TODAY:
            start_date = today
            end_date = today + timedelta(days=1)
            number_of_days = 1
        elif range_type == self.RANGE_TYPE_YESTERDAY:  # Yesterday
            start_date = today - timedelta(days=1)
            end_date = today
            number_of_days = 1
        elif range_type == self.RANGE_TYPE_TOMORROW:  # Tomorrow
            start_date = today + timedelta(days=1)
            end_date = start_date + timedelta(days=1)
            number_of_days = 1
        elif range_type == self.RANGE_TYPE_THIS_WEEK:  # This Week
            start_date = start_of_this_week
            end_date = start_date + timedelta(days=7)
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_NEXT_WEEK:  # Next Week
            start_date = start_of_this_week + timedelta(days=7)
            end_date = start_date + timedelta(days=7)
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_LAST_WEEK:  # Last Week
            start_date = start_of_this_week - timedelta(days=7)
            end_date = start_date + timedelta(days=7)
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_2_WEEKS_TIME:  # 2 Weeks Time
            start_date = start_of_this_week + timedelta(days=14)
            end_date = start_date + timedelta(days=7)
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_2_WEEKS_AGO:  # 2 Weeks Ago
            start_date = start_of_this_week - timedelta(days=14)
            end_date = start_date + timedelta(days=7)
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_3_WEEKS_TIME:  # 3 Weeks Time
            start_date = start_of_this_week + timedelta(days=21)
            end_date = start_date + timedelta(days=7)
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_3_WEEKS_AGO:  # 3 Weeks Ago
            start_date = start_of_this_week - timedelta(days=21)
            end_date = start_date + timedelta(days=7)
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_4_WEEKS_TIME:  # 4 Weeks Time
            start_date = start_of_this_week + timedelta(days=28)
            end_date = start_date + timedelta(days=7)
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_4_WEEKS_AGO:  # 4 Weeks Ago
            start_date = start_of_this_week - timedelta(days=28)
            end_date = start_date + timedelta(days=7)
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_5_WEEKS_TIME:  # 5 Weeks Time
            start_date = start_of_this_week + timedelta(days=35)
            end_date = start_date + timedelta(days=7)
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_5_WEEKS_AGO:  # 5 Weeks Ago
            start_date = start_of_this_week - timedelta(days=35)
            end_date = start_date + timedelta(days=7)
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_6_WEEKS_TIME:  # 6 Weeks Time
            start_date = start_of_this_week + timedelta(days=42)
            end_date = start_date + timedelta(days=7)
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_6_WEEKS_AGO:  # 6 Weeks Ago
            start_date = start_of_this_week - timedelta(days=42)
            end_date = start_date + timedelta(days=7)
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_THIS_MONTH:  # This Month
            start_date = date(today.year, today.month, 1)
            number_of_days = monthrange(today.year, today.month)[1]
            end_date = start_date + timedelta(days=number_of_days - 1)
        elif range_type == self.RANGE_TYPE_NEXT_MONTH:  # Next Month
            number_of_days_in_this_month = monthrange(today.year, today.month)[1]
            start_date = date(today.year, today.month, 1) + timedelta(days=number_of_days_in_this_month)
            number_of_days = monthrange(start_date.year, start_date.month)[1]
            end_date = start_date + timedelta(days=number_of_days)
        elif range_type == self.RANGE_TYPE_LAST_MONTH:  # Last Month
            end_date = date(today.year, today.month, 1) - timedelta(days=1)
            last_month = end_date
            start_date = date(last_month.year, last_month.month, 1)
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_NEXT_7_DAYS:  # Next 7 Days
            start_date = today
            end_date = start_date + timedelta(days=7)
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_LAST_7_DAYS:  # Last 7 Days
            start_date = today - timedelta(days=6)
            end_date = today
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_NEXT_14_DAYS:  # Next 14 Days
            start_date = today
            end_date = start_date + timedelta(days=14)
            number_of_days = 14
        elif range_type == self.RANGE_TYPE_LAST_14_DAYS:  # Last 14 Days
            start_date = today - timedelta(days=14)
            end_date = today
            number_of_days = 14
        elif range_type == self.RANGE_TYPE_NEXT_28_DAYS:  # Next 28 Days
            start_date = today
            end_date = start_date + timedelta(days=28)
            number_of_days = 28
        elif range_type == self.RANGE_TYPE_LAST_28_DAYS:  # Last 28 Days
            start_date = today - timedelta(days=28)
            end_date = today
            number_of_days = 28
        elif range_type == self.RANGE_TYPE_NEXT_60_DAYS:  # Next 60 Days
            start_date = today
            end_date = start_date + timedelta(days=60)
            number_of_days = 60
        elif range_type == self.RANGE_TYPE_LAST_60_DAYS:  # Last 60 Days
            start_date = today - timedelta(days=60)
            end_date = today
            number_of_days = 60
        elif range_type == self.RANGE_TYPE_NEXT_90_DAYS:  # Next 90 Days
            start_date = today
            end_date = start_date + timedelta(days=90)
            number_of_days = 90
        elif range_type == self.RANGE_TYPE_LAST_90_DAYS:  # Last 90 Days
            start_date = today - timedelta(days=90)
            end_date = today
            number_of_days = 90
        elif range_type == self.RANGE_TYPE_THIS_YEAR:  # This Year
            start_date = date(today.year, 1, 1)
            end_date = date(today.year, 12, 31)
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_LAST_YEAR:  # Last Year
            last_year = today.year - 1
            start_date = date(last_year, 1, 1)
            end_date = date(last_year, 12, 31)
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_LAST_18_MONTHS:
            date_offset = DateOffset()
            start_date = date_offset.get_offset('-18m', today)
            end_date = today
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_LAST_2_MONTHS:
            date_offset = DateOffset()
            start_date = date_offset.get_offset('-2m', today)
            end_date = today
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_NEXT_2_MONTHS:
            date_offset = DateOffset()
            start_date = today
            end_date = date_offset.get_offset('2m', today)
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_LAST_3_MONTHS:
            date_offset = DateOffset()
            start_date = date_offset.get_offset('-3m', today)
            end_date = today
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_NEXT_3_MONTHS:
            date_offset = DateOffset()
            start_date = today
            end_date = date_offset.get_offset('3m', today)
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_LAST_6_MONTHS:
            date_offset = DateOffset()
            start_date = date_offset.get_offset('-6m', today)
            end_date = today
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_NEXT_6_MONTHS:
            date_offset = DateOffset()
            start_date = today
            end_date = date_offset.get_offset('6m', today)
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_LAST_12_MONTHS:
            date_offset = DateOffset()
            start_date = date_offset.get_offset('-12m', today)
            end_date = today
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_NEXT_12_MONTHS:
            date_offset = DateOffset()
            start_date = today
            end_date = date_offset.get_offset('12m', today)
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_LAST_24_MONTHS:
            date_offset = DateOffset()
            start_date = date_offset.get_offset('-24m', today)
            end_date = today
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_NEXT_24_MONTHS:
            date_offset = DateOffset()
            start_date = today
            end_date = date_offset.get_offset('24m', today)
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_LAST_36_MONTHS:
            date_offset = DateOffset()
            start_date = date_offset.get_offset('-36m', today)
            end_date = today
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_NEXT_36_MONTHS:
            date_offset = DateOffset()
            start_date = today
            end_date = date_offset.get_offset('36m', today)
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_LAST_FINANCIAL_YEAR:
            fy_year = self.get_fy_start_year(today, financial_year_start_month, offset=-1)
            start_date, end_date = self._get_financial_year_bounds(fy_year, financial_year_start_month)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_THIS_FINANCIAL_YEAR:
            fy_year = self.get_fy_start_year(today, financial_year_start_month, offset=0)
            start_date, end_date = self._get_financial_year_bounds(fy_year, financial_year_start_month)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_NEXT_FINANCIAL_YEAR:
            fy_year = self.get_fy_start_year(today, financial_year_start_month, offset=1)
            start_date, end_date = self._get_financial_year_bounds(fy_year, financial_year_start_month)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_LAST_FINANCIAL_YEAR_Q1:
            fy_year = self.get_fy_start_year(today, financial_year_start_month, offset=-1)
            fy_start, _ = self._get_financial_year_bounds(fy_year, financial_year_start_month)
            start_date, end_date = self._get_financial_quarter_bounds(fy_start, 1)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_LAST_FINANCIAL_YEAR_Q2:
            fy_year = self.get_fy_start_year(today, financial_year_start_month, offset=-1)
            fy_start, _ = self._get_financial_year_bounds(fy_year, financial_year_start_month)
            start_date, end_date = self._get_financial_quarter_bounds(fy_start, 2)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_LAST_FINANCIAL_YEAR_Q3:
            fy_year = self.get_fy_start_year(today, financial_year_start_month, offset=-1)
            fy_start, _ = self._get_financial_year_bounds(fy_year, financial_year_start_month)
            start_date, end_date = self._get_financial_quarter_bounds(fy_start, 3)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_LAST_FINANCIAL_YEAR_Q4:
            fy_year = self.get_fy_start_year(today, financial_year_start_month, offset=-1)
            fy_start, _ = self._get_financial_year_bounds(fy_year, financial_year_start_month)
            start_date, end_date = self._get_financial_quarter_bounds(fy_start, 4)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_THIS_FINANCIAL_YEAR_Q1:
            fy_year = self.get_fy_start_year(today, financial_year_start_month, offset=0)
            fy_start, _ = self._get_financial_year_bounds(fy_year, financial_year_start_month)
            start_date, end_date = self._get_financial_quarter_bounds(fy_start, 1)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_THIS_FINANCIAL_YEAR_Q2:
            fy_year = self.get_fy_start_year(today, financial_year_start_month, offset=0)
            fy_start, _ = self._get_financial_year_bounds(fy_year, financial_year_start_month)
            start_date, end_date = self._get_financial_quarter_bounds(fy_start, 2)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_THIS_FINANCIAL_YEAR_Q3:
            fy_year = self.get_fy_start_year(today, financial_year_start_month, offset=0)
            fy_start, _ = self._get_financial_year_bounds(fy_year, financial_year_start_month)
            start_date, end_date = self._get_financial_quarter_bounds(fy_start, 3)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_THIS_FINANCIAL_YEAR_Q4:
            fy_year = self.get_fy_start_year(today, financial_year_start_month, offset=0)
            fy_start, _ = self._get_financial_year_bounds(fy_year, financial_year_start_month)
            start_date, end_date = self._get_financial_quarter_bounds(fy_start, 4)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_NEXT_FINANCIAL_YEAR_Q1:
            fy_year = self.get_fy_start_year(today, financial_year_start_month, offset=1)
            fy_start, _ = self._get_financial_year_bounds(fy_year, financial_year_start_month)
            start_date, end_date = self._get_financial_quarter_bounds(fy_start, 1)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_NEXT_FINANCIAL_YEAR_Q2:
            fy_year = self.get_fy_start_year(today, financial_year_start_month, offset=1)
            fy_start, _ = self._get_financial_year_bounds(fy_year, financial_year_start_month)
            start_date, end_date = self._get_financial_quarter_bounds(fy_start, 2)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_NEXT_FINANCIAL_YEAR_Q3:
            fy_year = self.get_fy_start_year(today, financial_year_start_month, offset=1)
            fy_start, _ = self._get_financial_year_bounds(fy_year, financial_year_start_month)
            start_date, end_date = self._get_financial_quarter_bounds(fy_start, 3)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_NEXT_FINANCIAL_YEAR_Q4:
            fy_year = self.get_fy_start_year(today, financial_year_start_month, offset=1)
            fy_start, _ = self._get_financial_year_bounds(fy_year, financial_year_start_month)
            start_date, end_date = self._get_financial_quarter_bounds(fy_start, 4)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_LAST_FINANCIAL_QUARTER:
            fy_year, quarter = self._get_current_fy_quarter(today, financial_year_start_month)

            # Move back one quarter
            if quarter == 1:
                quarter = 4
                fy_year -= 1
            else:
                quarter -= 1

            fy_start, _ = self._get_financial_year_bounds(fy_year, financial_year_start_month)
            start_date, end_date = self._get_financial_quarter_bounds(fy_start, quarter)

            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_CURRENT_FINANCIAL_QUARTER:
            fy_year, quarter = self._get_current_fy_quarter(today, financial_year_start_month)

            fy_start, _ = self._get_financial_year_bounds(fy_year, financial_year_start_month)
            start_date, end_date = self._get_financial_quarter_bounds(fy_start, quarter)

            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_NEXT_FINANCIAL_QUARTER:
            fy_year, quarter = self._get_current_fy_quarter(today, financial_year_start_month)

            # Move forward one quarter
            if quarter == 4:
                quarter = 1
                fy_year += 1
            else:
                quarter += 1

            fy_start, _ = self._get_financial_year_bounds(fy_year, financial_year_start_month)
            start_date, end_date = self._get_financial_quarter_bounds(fy_start, quarter)

            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_LAST_YEAR_SAME_FINANCIAL_QUARTER:
            # Determine current quarter in THIS FY
            this_fy_year, current_quarter = self._get_current_fy_quarter(today, financial_year_start_month)

            # Move back one FY but keep same quarter number
            fy_year = this_fy_year - 1

            fy_start, _ = self._get_financial_year_bounds(fy_year, financial_year_start_month)
            start_date, end_date = self._get_financial_quarter_bounds(fy_start, current_quarter)

            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_LAST_CALENDAR_QUARTER:
            # Current calendar quarter
            quarter = self._get_calendar_quarter(today)
            year = today.year

            # Move back one quarter
            if quarter == 1:
                quarter = 4
                year -= 1
            else:
                quarter -= 1

            start_date, end_date = self._get_calendar_quarter_bounds(year, quarter)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_THIS_CALENDAR_QUARTER:
            quarter = self._get_calendar_quarter(today)
            year = today.year

            start_date, end_date = self._get_calendar_quarter_bounds(year, quarter)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_NEXT_CALENDAR_QUARTER:
            quarter = self._get_calendar_quarter(today)
            year = today.year

            # Move forward one quarter
            if quarter == 4:
                quarter = 1
                year += 1
            else:
                quarter += 1

            start_date, end_date = self._get_calendar_quarter_bounds(year, quarter)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_LAST_CALENDAR_YEAR_Q1:
            year = today.year - 1
            start_date, end_date = self._get_calendar_quarter_bounds(year, 1)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_LAST_CALENDAR_YEAR_Q2:
            year = today.year - 1
            start_date, end_date = self._get_calendar_quarter_bounds(year, 2)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_LAST_CALENDAR_YEAR_Q3:
            year = today.year - 1
            start_date, end_date = self._get_calendar_quarter_bounds(year, 3)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_LAST_CALENDAR_YEAR_Q4:
            year = today.year - 1
            start_date, end_date = self._get_calendar_quarter_bounds(year, 4)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_THIS_CALENDAR_YEAR_Q1:
            year = today.year
            start_date, end_date = self._get_calendar_quarter_bounds(year, 1)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_THIS_CALENDAR_YEAR_Q2:
            year = today.year
            start_date, end_date = self._get_calendar_quarter_bounds(year, 2)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_THIS_CALENDAR_YEAR_Q3:
            year = today.year
            start_date, end_date = self._get_calendar_quarter_bounds(year, 3)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_THIS_CALENDAR_YEAR_Q4:
            year = today.year
            start_date, end_date = self._get_calendar_quarter_bounds(year, 4)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_NEXT_CALENDAR_YEAR_Q1:
            year = today.year + 1
            start_date, end_date = self._get_calendar_quarter_bounds(year, 1)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_NEXT_CALENDAR_YEAR_Q2:
            year = today.year + 1
            start_date, end_date = self._get_calendar_quarter_bounds(year, 2)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_NEXT_CALENDAR_YEAR_Q3:
            year = today.year + 1
            start_date, end_date = self._get_calendar_quarter_bounds(year, 3)
            number_of_days = (end_date - start_date).days + 1

        elif range_type == self.RANGE_TYPE_NEXT_CALENDAR_YEAR_Q4:
            year = today.year + 1
            start_date, end_date = self._get_calendar_quarter_bounds(year, 4)
            number_of_days = (end_date - start_date).days + 1

        else:
            raise AssertionError('unknown date value')

        start_date_and_time = datetime.combine(start_date, datetime.min.time())
        end_date_and_time = datetime.combine(end_date, datetime.max.time())

        return start_date_and_time, end_date_and_time, number_of_days

    def get_variable_date_filter_values(self):
        values = {}
        for choice in self.RANGE_TYPE_CHOICES:
            values['#variable_date:%d' % choice[0]] = choice[1]
        return values

    @staticmethod
    def get_date_filter_years():
        values = {}
        today = date.today()
        start_year = today.year - 10
        end_year = start_year + 15

        for year in range(start_year, end_year):
            values['#year:%d' % year] = year
        return values

    @staticmethod
    def get_date_filter_months():
        values = {}
        for index, month in enumerate(list(calendar.month_name)[1:], 1):
            values['#month:%d' % index] = month
        return values

    @staticmethod
    def get_date_filter_quarters():
        values = {}
        for quarter in range(1, 5):
            values['#quarter:%d' % quarter] = f'Quarter {quarter}'
        for quarter in range(1, 5):
            values['#financial_quarter:%d' % quarter] = f'Financial Quarter {quarter}'
        return values

    @staticmethod
    def get_fy_start_year(today, fy_start_month, offset=0):
        """
        Returns the start year of the financial year.
        offset = -1 (last FY), 0 (this FY), +1 (next FY)
        """
        this_fy_start_year = today.year if today.month >= fy_start_month else today.year - 1
        return this_fy_start_year + offset

    @staticmethod
    def _get_financial_year_bounds(year, start_month):
        """
        Returns inclusive financial year date range:
        start_date, end_date
        """
        start = date(year, start_month, 1)

        # start of next FY
        next_start = date(year + 1, 1, 1) if start_month == 1 else date(year + 1, start_month, 1)

        end = next_start - timedelta(days=1)
        return start, end

    @staticmethod
    def _get_financial_quarter_bounds(fy_start_date, quarter_number):
        """
        Returns inclusive quarter range inside a financial year.
        """
        q_start = fy_start_date + relativedelta(months=3 * (quarter_number - 1))
        q_end = q_start + relativedelta(months=3) - timedelta(days=1)
        return q_start, q_end

    def _get_current_fy_quarter(self, today, fy_start_month):
        # Get start of this FY
        fy_year = self.get_fy_start_year(today, fy_start_month, offset=0)
        fy_start, _ = self._get_financial_year_bounds(fy_year, fy_start_month)

        # Months difference from FY start
        delta_months = (today.year - fy_start.year) * 12 + (today.month - fy_start.month)

        # Quarter index (0,1,2,3)
        quarter = delta_months // 3 + 1
        return fy_year, quarter

    @staticmethod
    def _get_calendar_quarter(dt):
        return (dt.month - 1) // 3 + 1

    def _get_calendar_quarter_bounds(self, year, quarter):
        """
        Calendar quarter is simply a financial quarter where FY starts in January.
        """
        fy_start_date = date(year, 1, 1)
        return self._get_financial_quarter_bounds(fy_start_date, quarter)



from django.db.models import Count, ExpressionWrapper, FloatField, Sum
from django.db.models.functions import Coalesce, NullIf

from advanced_report_builder.column_types import NUMBER_FIELDS
from advanced_report_builder.exceptions import ReportError
from advanced_report_builder.globals import (
    ANNOTATION_CHOICE_SUM,
)
from advanced_report_builder.views.charts_base import ChartBaseView


class ValueBaseView(ChartBaseView):
    use_annotations = False

    def _get_count(self, fields):
        number_function_kwargs = {
            'aggregations': {'count': Count(1)},
            'field': 'count',
            'column_name': 'count',
            'options': {'calculated': True},
            'model_path': '',
        }

        field = self.number_field(**number_function_kwargs)
        fields.append(field)

    def _process_aggregations(
        self,
        field,
        base_model,
        report_builder_class,
        decimal_places,
        fields,
        aggregations_type=ANNOTATION_CHOICE_SUM,
        divider=None,
    ):
        django_field, col_type_override, _, _ = self.get_field_details(
            base_model=base_model,
            field=field,
            report_builder_class=report_builder_class,
        )

        if isinstance(django_field, NUMBER_FIELDS) or col_type_override is not None and col_type_override.annotations:
            self.get_number_field(
                annotations_type=aggregations_type,
                append_annotation_query=False,
                index=0,
                data_attr={},
                table_field={'field': field, 'title': field},
                fields=fields,
                col_type_override=col_type_override,
                decimal_places=decimal_places,
                convert_currency_fields=True,
                divider=divider,
            )
        else:
            raise ReportError('not a number field')

    def get_percentage_field(
        self,
        fields,
        numerator_field_name,
        numerator_col_type_override,
        denominator_field_name,
        denominator_col_type_override,
        numerator_filter,
        decimal_places=0,
    ):
        if numerator_col_type_override:
            actual_numerator_field_name = numerator_col_type_override.field
        else:
            actual_numerator_field_name = numerator_field_name

        if denominator_col_type_override:
            actual_denominator_field_name = denominator_col_type_override.field
        else:
            actual_denominator_field_name = denominator_field_name

        number_function_kwargs = {}
        if decimal_places:
            number_function_kwargs = {'decimal_places': int(decimal_places)}

        new_field_name = f'{actual_numerator_field_name}_{actual_denominator_field_name}'

        if numerator_col_type_override is not None and numerator_col_type_override.annotations:
            if not isinstance(numerator_col_type_override.annotations, dict):
                raise ReportError('Unknown annotation type')

            annotations = list(numerator_col_type_override.annotations.values())[0]
            if numerator_filter and isinstance(annotations, Sum | Count):
                annotations.filter = numerator_filter
                numerator = Coalesce(annotations + 0.0, 0.0)
            else:
                numerator = Coalesce(annotations + 0.0, 0.0)
        else:
            if numerator_filter:
                numerator = Coalesce(
                    (Sum(actual_numerator_field_name, filter=numerator_filter) + 0.0),
                    0.0,
                )
            else:
                numerator = Coalesce((Sum(actual_numerator_field_name) + 0.0), 0.0)

        if denominator_col_type_override is not None and denominator_col_type_override.annotations:
            if not isinstance(denominator_col_type_override.annotations, dict):
                raise ReportError('Unknown annotation type')

            annotations = list(denominator_col_type_override.annotations.values())[0]

            denominator = Coalesce(annotations + 0.0, 0.0)
        else:
            denominator = Coalesce(Sum(actual_denominator_field_name) + 0.0, 0.0)

        number_function_kwargs['aggregations'] = {
            new_field_name: ExpressionWrapper((numerator / NullIf(denominator, 0)) * 100.0, output_field=FloatField())
        }
        field_name = new_field_name

        number_function_kwargs.update(
            {
                'field': field_name,
                'column_name': field_name,
                'options': {'calculated': True},
                'model_path': '',
            }
        )

        field = self.number_field(**number_function_kwargs)

        fields.append(field)

        return field_name

    def _process_percentage(
        self, denominator_field, numerator_field, base_model, report_builder_class, decimal_places, fields
    ):
        if denominator_field is None:
            raise ReportError('denominator field is None')
        if numerator_field is None:
            raise ReportError('numerator field is None')
        deno_django_field, denominator_col_type_override, _, _ = self.get_field_details(
            base_model=base_model,
            field=denominator_field,
            report_builder_class=report_builder_class,
        )
        if not isinstance(deno_django_field, NUMBER_FIELDS) and (
            denominator_col_type_override is not None and not denominator_col_type_override.annotations
        ):
            raise ReportError('denominator is not a number field')

        num_django_field, numerator_col_type_override, _, _ = self.get_field_details(
            base_model=base_model,
            field=numerator_field,
            report_builder_class=report_builder_class,
        )

        if not isinstance(num_django_field, NUMBER_FIELDS) and (
            numerator_col_type_override is not None and not numerator_col_type_override.annotations
        ):
            raise ReportError('numerator is not a number field')

        numerator_filter = None
        report_query = self.get_report_query(report=self.chart_report)
        if report_query:
            numerator_filter = self.process_filters(search_filter_data=report_query.extra_query)

        self.get_percentage_field(
            fields=fields,
            numerator_field_name=numerator_field,
            numerator_col_type_override=numerator_col_type_override,
            denominator_field_name=denominator_field,
            denominator_col_type_override=denominator_col_type_override,
            numerator_filter=numerator_filter,
            decimal_places=decimal_places,
        )

    def _process_percentage_from_count(self, numerator_filter, decimal_places, fields):
        if numerator_filter:
            numerator = Coalesce(Count(1, filter=numerator_filter) + 0.0, 0.0)
        else:
            numerator = Coalesce(Count(1) + 0.0, 0.0)
        denominator = Coalesce(Count(1) + 0.0, 0.0)
        a = (numerator / denominator) * 100.0

        number_function_kwargs = {
            'aggregations': {'count': a},
            'field': 'count',
            'column_name': 'count',
            'options': {'calculated': True},
            'model_path': '',
        }
        if decimal_places:
            number_function_kwargs['decimal_places'] = int(decimal_places)

        field = self.number_field(**number_function_kwargs)
        fields.append(field)



from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.forms import ChoiceField
from django.utils.dates import MONTHS
from django_datatables.columns import DatatableColumn, ManyToManyColumn, NoHeadingColumn
from django_datatables.model_def import DatatableModel
from django_modals.model_fields.colour import ColourField
from django_modals.widgets.select2 import Select2
from time_stamped_model.models import TimeStampedModel

from advanced_report_builder.globals import (
    ANNOTATION_CHART_SCALE,
    ANNOTATION_CHOICE_COUNT,
    ANNOTATION_VALUE_CHOICES,
    ANNOTATIONS_CHOICES,
    CALENDAR_VIEW_TYPE_CHOICES,
    CALENDAR_VIEW_TYPE_DAY,
    CALENDAR_VIEW_TYPE_GRID_WEEK,
    CALENDAR_VIEW_TYPE_LIST_WEEK,
    CALENDAR_VIEW_TYPE_MONTH,
    CALENDAR_VIEW_TYPE_YEAR,
    DISPLAY_OPTION_2_PER_ROW,
    DISPLAY_OPTION_CHOICES,
    DISPLAY_OPTION_CLASSES,
    DISPLAY_OPTION_NONE,
)
from advanced_report_builder.signals import model_report_save


class Target(TimeStampedModel):
    class TargetType(models.IntegerChoices):
        COUNT = 1, 'Count'
        MONEY = 2, 'Money'
        PERCENTAGE = 3, 'Percentage'

    class PeriodType(models.IntegerChoices):
        DAILY = 1, 'Daily'
        WEEKLY = 2, 'Weekly'
        MONTHLY = 3, 'Monthly'
        QUARTER = 4, 'Quarterly'

    slug = models.SlugField(unique=True)
    name = models.CharField(max_length=64)
    target_type = models.PositiveSmallIntegerField(choices=TargetType.choices)
    period_type = models.PositiveSmallIntegerField(choices=PeriodType.choices, default=PeriodType.MONTHLY)
    colour = ColourField(null=True, blank=True, help_text='The colour when it gets displayed on a report')
    default_value = models.IntegerField(blank=True, null=True)
    default_percentage = models.FloatField(blank=True, null=True)
    overridden = models.BooleanField(default=False)
    override_data = models.JSONField(null=True, blank=True)

    def get_value(self):
        if self.target_type == self.TargetType.PERCENTAGE:
            return self.default_percentage
        return self.default_value

    def __str__(self):
        return self.name

    def get_override_data(self):
        override_data = self.override_data if self.override_data else {}

        ordered_value_dict = {}

        for year, data in override_data.items():
            if not ordered_value_dict.get(year):
                ordered_value_dict[year] = {}
            for month in MONTHS.values():
                if data.get(str(month)):
                    ordered_value_dict[year][month] = data[month]
                else:
                    ordered_value_dict[year][month] = self.default_value

        return ordered_value_dict

    def add_year(self, year):
        override_data = self.override_data if self.override_data else {}
        if not override_data.get(str(year)):
            override_data[str(year)] = {}
        temp = {}
        # Create a temp dict that has the keys as integers.
        for key, value in override_data.items():
            temp[int(key)] = value
        sorted_dict = {}
        for key in sorted(temp):
            sorted_dict[str(key)] = temp[key]
        if sorted_dict:
            self.override_data = sorted_dict
            self.save()

    def save(self, *args, **kwargs):
        """
        Overrides the save so that a new slug is generated
        :param args:
        :param kwargs:
        """
        self.make_new_slug()
        super().save(*args, **kwargs)


class ReportTag(TimeStampedModel):
    name = models.CharField(max_length=128, unique=True)
    slug = models.SlugField(unique=True)
    order = models.PositiveSmallIntegerField()

    class Meta:
        ordering = ('order',)

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        self.make_new_slug(allow_dashes=False, on_edit=True)
        self.set_order_field()
        return super().save(*args, **kwargs)


class ReportType(TimeStampedModel):
    name = models.CharField(max_length=200)
    slug = models.SlugField(unique=True)
    content_type = models.ForeignKey(ContentType, null=False, blank=False, on_delete=models.PROTECT)
    report_builder_class_name = models.CharField(max_length=200)

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']

    def save(self, *args, **kwargs):
        self.make_new_slug(allow_dashes=False, on_edit=True)
        return super().save(*args, **kwargs)


class Report(TimeStampedModel):
    output_types = {
        'tablereport': 'Table',
        'singlevaluereport': 'Single Value',
        'barchartreport': 'Bar Chart',
        'linechartreport': 'Line Chart',
        'piechartreport': 'Pie Chart',
        'funnelchartreport': 'Funnel Chart',
        'kanbanreport': 'Kanban',
        'customreport': 'Custom',
        'calendarreport': 'Calendar',
        'multivaluereport': 'Multi Values',
    }

    output_types_icons = {
        'tablereport': '<i class="fas fa-table"></i>',
        'singlevaluereport': '<i class="fas fa-box-open"></i>',
        'barchartreport': '<i class="fas fa-chart-bar"></i>',
        'linechartreport': '<i class="fas fa-chart-line"></i>',
        'piechartreport': '<i class="fas fa-chart-pie"></i>',
        'funnelchartreport': '<i class="fas fa-filter"></i>',
        'kanbanreport': '<i class="fas fa-chart-bar fa-flip-vertical"></i>',
        'customreport': '<i class="fas fa-file"></i>',
        'calendarreport': '<i class="fas fa-calendar"></i>',
        'multivaluereport': '<i class="fas fa-grip-horizontal"></i>',
    }

    name = models.CharField(max_length=200)
    slug = models.SlugField(unique=True)
    slug_alias = models.SlugField(blank=True, null=True)  # used if the slug changes
    report_type = models.ForeignKey(ReportType, null=True, blank=False, on_delete=models.PROTECT)
    instance_type = models.CharField(null=True, max_length=255)
    report_tags = models.ManyToManyField(ReportTag, blank=True)
    notes = models.TextField(null=True, blank=True)
    version = models.PositiveSmallIntegerField(default=0)
    user_created = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name='report_user_created_set',
    )
    user_updated = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name='report_user_updated_set',
    )

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']

    def get_report(self):
        return getattr(self, self.instance_type)

    def get_title(self):
        return self.name

    def get_output_type_name(self):
        return self.output_types[self.instance_type]

    def save(self, *args, **kwargs):
        current_user = getattr(self, '_current_user', None)
        if current_user is not None and current_user.is_authenticated:
            self.user_updated = current_user
            if self.user_created is None:
                self.user_created = current_user
        is_new = self.pk is None
        if self.version is not None:
            self.version += 1
        slug_alias = self.slug
        self.make_new_slug(obj=Report, allow_dashes=False, on_edit=True)
        if slug_alias != self.slug:
            self.slug_alias = slug_alias
        if self.instance_type is None:
            self.instance_type = self._meta.label_lower.split('.')[1]
        result = super().save(*args, **kwargs)
        model_report_save.send(sender=self.__class__, instance=self, created=is_new, user=current_user)
        return result

    def get_base_model(self):
        if self.report_type is None:
            return None
        return self.report_type.content_type.model_class()

    def show_dashboard_query(self):
        return True

    def dashboard_fields(self, form, dashboard_report):
        pass

    def save_extra_dashboard_fields(self, form, dashboard_report):
        pass

    class Datatable(DatatableModel):
        class OutputType(DatatableColumn):
            def col_setup(self):
                self.field = ['instance_type', 'customreport__output_type']

            # noinspection PyMethodMayBeStatic
            def row_result(self, data, _page_data):
                instance_type = data[self.model_path + 'instance_type']
                if instance_type == 'customreport':
                    output_type = data[self.model_path + 'customreport__output_type']
                    if output_type:
                        return output_type
                return Report.output_types.get(instance_type, '')

        class OutputTypeIcon(NoHeadingColumn):
            def col_setup(self):
                self.field = ['instance_type']

            # noinspection PyMethodMayBeStatic
            def row_result(self, data, _page_data):
                instance_type = data[self.model_path + 'instance_type']
                return Report.output_types_icons.get(instance_type, '')

        report_tags_badge = ManyToManyColumn(
            field='report_tags__name',
            html='<span class="badge badge-primary"> %1% </span>',
            title='Tags',
        )


class ReportQuery(TimeStampedModel):
    report = models.ForeignKey(Report, on_delete=models.CASCADE)
    name = models.CharField(max_length=256, default='Standard')
    query = models.JSONField(null=True, blank=True)
    extra_query = models.JSONField(null=True, blank=True)  # used for single value Numerator
    target = models.ForeignKey(Target, blank=True, null=True, on_delete=models.SET_NULL)
    order = models.PositiveSmallIntegerField()

    def save(self, *args, **kwargs):
        self.set_order_field(extra_filters={'report': self.report})
        result = super().save(*args, **kwargs)
        self.report._current_user = getattr(self, '_current_user', None)  # this is to update the users
        self.report.save()
        return result

    class Meta:
        ordering = ['order']
        verbose_name_plural = 'Report queries'

    def __str__(self):
        return self.name


class ReportQueryOrder(TimeStampedModel):
    report_query = models.ForeignKey(ReportQuery, on_delete=models.CASCADE)
    order_by_field = models.CharField(max_length=200, blank=True, null=True)
    order_by_ascending = models.BooleanField(default=True)
    order = models.PositiveSmallIntegerField()

    class Meta:
        ordering = ['order']

    def save(self, *args, **kwargs):
        self.set_order_field(extra_filters={'report_query': self.report_query})
        return super().save(*args, **kwargs)


class TableReport(Report):
    table_fields = models.JSONField(null=True, blank=True)
    has_clickable_rows = models.BooleanField(default=False)
    link_field = models.CharField(max_length=200, blank=True, null=True)
    pivot_fields = models.JSONField(null=True, blank=True)
    order_by_field = models.CharField(max_length=200, blank=True, null=True)
    order_by_ascending = models.BooleanField(default=True)
    page_length = models.PositiveSmallIntegerField(
        choices=(
            (10, '10'),
            (25, '25'),
            (50, '50'),
            (100, '100'),
            (150, '150'),
            (200, '200'),
        ),
        default=100,
    )


class SingleValueReport(Report):
    class SingleValueType(models.IntegerChoices):
        COUNT = 1, 'Count'
        SUM = 2, 'Sum'
        COUNT_AND_SUM = 3, 'Count & Sum'
        PERCENT = 4, 'Percent'
        PERCENT_FROM_COUNT = 5, 'Percent from Count'
        AVERAGE_SUM_FROM_COUNT = 6, 'Average Sum from Count'
        AVERAGE_SUM_OVER_TIME = 7, 'Average Sum over Time'
        AVERAGE_SUM_OVER_TIME_EXCLUDING_WEEKENDS = 8, 'Average Sum over Time (Excluding Weekends)'

        @classmethod
        def is_percentage(cls, value):
            return value in {
                cls.PERCENT,
                cls.PERCENT_FROM_COUNT,
            }

    tile_colour = ColourField(blank=True, null=True)
    field = models.CharField(max_length=200, blank=True, null=True)  # denominator
    numerator = models.CharField(max_length=200, blank=True, null=True)
    single_value_type = models.PositiveSmallIntegerField(choices=SingleValueType.choices, default=SingleValueType.COUNT)
    prefix = models.CharField(max_length=64, blank=True, null=True)
    decimal_places = models.IntegerField(default=0)

    show_breakdown = models.BooleanField(default=False)
    breakdown_fields = models.JSONField(null=True, blank=True)

    average_scale = models.PositiveSmallIntegerField(choices=ANNOTATION_VALUE_CHOICES, blank=True, null=True)
    average_start_period = models.PositiveSmallIntegerField(blank=True, null=True)
    average_end_period = models.PositiveSmallIntegerField(blank=True, null=True)

    def is_percentage(self):
        return self.SingleValueType.is_percentage(self.single_value_type)


class BarChartReport(Report):
    class BarChartOrientation(models.IntegerChoices):
        VERTICAL = 1, 'Vertical'
        HORIZONTAL = 2, 'Horizontal'

    DATE_FIELD_SINGLE = 1
    DATE_FIELD_RANGE = 2

    DATE_FIELD_CHOICES = ((DATE_FIELD_SINGLE, 'Single'), (DATE_FIELD_RANGE, 'Range'))

    axis_scale = models.PositiveSmallIntegerField(choices=ANNOTATION_VALUE_CHOICES)

    date_field_type = models.PositiveSmallIntegerField(choices=DATE_FIELD_CHOICES, default=DATE_FIELD_SINGLE)

    date_field = models.CharField(max_length=200)
    end_date_field = models.CharField(max_length=200, blank=True, null=True)

    axis_value_type = models.PositiveSmallIntegerField(
        choices=ANNOTATIONS_CHOICES,
        default=ANNOTATION_CHOICE_COUNT,
        null=True,
        blank=True,
    )
    fields = models.JSONField(null=True, blank=True)
    x_label = models.CharField(max_length=200, blank=True, null=True)
    y_label = models.CharField(max_length=200, blank=True, null=True)

    bar_chart_orientation = models.PositiveSmallIntegerField(
        choices=BarChartOrientation.choices, default=BarChartOrientation.VERTICAL
    )
    stacked = models.BooleanField(default=False)
    show_totals = models.BooleanField(default=False)
    show_blank_dates = models.BooleanField(default=True)

    show_breakdown = models.BooleanField(default=False)
    breakdown_fields = models.JSONField(null=True, blank=True)

    def is_orientation_vertical(self):
        return self.bar_chart_orientation == self.BarChartOrientation.VERTICAL

    def get_chart_scale(self):
        return ANNOTATION_CHART_SCALE[self.axis_scale]


class LineChartReport(Report):
    axis_scale = models.PositiveSmallIntegerField(choices=ANNOTATION_VALUE_CHOICES)
    date_field = models.CharField(max_length=200)
    axis_value_type = models.PositiveSmallIntegerField(
        choices=ANNOTATIONS_CHOICES,
        default=ANNOTATION_CHOICE_COUNT,
        null=True,
        blank=True,
    )
    fields = models.JSONField(null=True, blank=True)
    x_label = models.CharField(max_length=200, blank=True, null=True)
    y_label = models.CharField(max_length=200, blank=True, null=True)
    show_totals = models.BooleanField(default=False)

    has_targets = models.BooleanField(default=False)
    targets = models.ManyToManyField(Target, blank=True)

    def get_chart_scale(self):
        return ANNOTATION_CHART_SCALE[self.axis_scale]


class PieChartReport(Report):
    class PieChartStyle(models.IntegerChoices):
        PIE = 1, 'Pie'
        DOUGHNUT = 2, 'Doughnut'

    axis_value_type = models.PositiveSmallIntegerField(
        choices=ANNOTATIONS_CHOICES,
        default=ANNOTATION_CHOICE_COUNT,
        null=True,
        blank=True,
    )
    fields = models.JSONField(null=True, blank=True)
    style = models.PositiveSmallIntegerField(choices=PieChartStyle.choices, default=PieChartStyle.PIE)

    def is_pie_chart(self):
        return self.style == self.PieChartStyle.PIE


class FunnelChartReport(Report):
    axis_value_type = models.PositiveSmallIntegerField(
        choices=ANNOTATIONS_CHOICES,
        default=ANNOTATION_CHOICE_COUNT,
        null=True,
        blank=True,
    )
    fields = models.JSONField(null=True, blank=True)


class KanbanReport(Report):
    def show_dashboard_query(self):
        return False  # show queries if true


class KanbanReportDescription(TimeStampedModel):
    kanban_report = models.ForeignKey(KanbanReport, on_delete=models.CASCADE)
    name = models.CharField(max_length=200)
    report_type = models.ForeignKey(ReportType, null=True, blank=False, on_delete=models.PROTECT)
    description = models.TextField(blank=True, null=True)
    order = models.PositiveSmallIntegerField()

    def save(self, *args, **kwargs):
        self.set_order_field(extra_filters={'kanban_report': self.kanban_report})
        result = super().save(*args, **kwargs)
        self.kanban_report._current_user = getattr(self, '_current_user', None)  # this is to update the users
        self.kanban_report.save()
        return result

    def get_base_model(self):
        return self.report_type.content_type.model_class()

    def __str__(self):
        return self.name

    class Meta:
        ordering = ('order',)


class KanbanReportLane(TimeStampedModel):
    MULTIPLE_TYPE_NA = 0
    MULTIPLE_TYPE_DAILY = 1
    MULTIPLE_TYPE_DAILY_WITHIN = 2
    MULTIPLE_TYPE_WEEKLY = 3
    MULTIPLE_TYPE_WEEKLY_WITHIN = 4
    MULTIPLE_TYPE_MONTHLY = 5
    MULTIPLE_TYPE_MONTHLY_WITHIN = 6

    MULTIPLE_TYPE_CHOICES = [
        (MULTIPLE_TYPE_NA, 'N/A'),
        (MULTIPLE_TYPE_DAILY, 'Daily (single date)'),
        (MULTIPLE_TYPE_DAILY_WITHIN, 'Daily (within two date)'),
        (MULTIPLE_TYPE_WEEKLY, 'Weekly (single date)'),
        (MULTIPLE_TYPE_WEEKLY_WITHIN, 'Weekly (within two date)'),
        (MULTIPLE_TYPE_MONTHLY, 'Monthly (single date)'),
        (MULTIPLE_TYPE_MONTHLY_WITHIN, 'Monthly (within two date)'),
    ]

    kanban_report = models.ForeignKey(KanbanReport, on_delete=models.CASCADE)
    name = models.CharField(max_length=200)
    order = models.PositiveSmallIntegerField()
    report_type = models.ForeignKey(ReportType, null=True, blank=False, on_delete=models.PROTECT)
    heading_field = models.CharField(max_length=200, blank=True, null=True)
    link_field = models.CharField(max_length=200, blank=True, null=True)
    order_by_field = models.CharField(max_length=200, blank=True, null=True)
    order_by_ascending = models.BooleanField(default=True)
    kanban_report_description = models.ForeignKey(
        KanbanReportDescription, null=True, blank=False, on_delete=models.CASCADE
    )

    multiple_type = models.PositiveIntegerField(choices=MULTIPLE_TYPE_CHOICES, default=MULTIPLE_TYPE_NA)
    multiple_type_label = models.CharField(max_length=200, blank=True, null=True)
    multiple_type_date_field = models.CharField(max_length=200, blank=True, null=True)
    multiple_type_end_date_field = models.CharField(max_length=200, blank=True, null=True)

    # this could be choice field from RANGE_TYPE_CHOICES however if one adds a new one it creates a new migration!
    multiple_start_period = models.PositiveSmallIntegerField(blank=True, null=True)
    multiple_end_period = models.PositiveSmallIntegerField(blank=True, null=True)
    background_colour_field = models.CharField(max_length=200, blank=True, null=True)
    heading_colour_field = models.CharField(max_length=200, blank=True, null=True)

    query_data = models.JSONField(null=True, blank=True)

    def save(self, *args, **kwargs):
        self.set_order_field(extra_filters={'kanban_report': self.kanban_report})
        result = super().save(*args, **kwargs)
        self.kanban_report._current_user = getattr(self, '_current_user', None)  # this is to update the users
        self.kanban_report.save()
        return result

    def get_base_model(self):
        return self.report_type.content_type.model_class()

    class Meta:
        ordering = ('order',)


class MultiValueReport(Report):
    rows = models.PositiveSmallIntegerField()
    columns = models.PositiveSmallIntegerField()
    default_multi_cell_style = models.ForeignKey('MultiCellStyle', on_delete=models.PROTECT, null=True, blank=True)


class MultiCellStyle(TimeStampedModel):
    class AlignType(models.IntegerChoices):
        LEFT = 0, 'Left'
        CENTRE = 1, 'Centre'
        RIGHT = 2, 'Right'

    multi_value_report = models.ForeignKey(MultiValueReport, on_delete=models.CASCADE)
    name = models.CharField(max_length=200)
    align_type = models.IntegerField(choices=AlignType.choices, default=AlignType.LEFT)
    bold = models.BooleanField(default=False)
    italic = models.BooleanField(default=False)
    font_size = models.PositiveSmallIntegerField(null=True, blank=True)
    font_colour = ColourField(null=True, blank=True)
    background_colour = ColourField(null=True, blank=True)

    def __str__(self):
        return self.name

    def get_td_class(self):
        results = []
        if self.align_type == self.AlignType.CENTRE:
            results.append('text-center')
        elif self.align_type == self.AlignType.RIGHT:
            results.append('text-right')
        if self.bold:
            results.append('font-weight-bold')
        if self.italic:
            results.append('font-italic')
        return ' '.join(results)

    def get_td_style(self):
        results = []
        if self.font_size:
            results.append(f'font-size: {self.font_size}px;')
        if self.font_colour:
            results.append(f'color: #{self.font_colour};')
        if self.background_colour:
            results.append(f'background-color: #{self.background_colour};')
        return ';'.join(results)


class MultiValueHeldQuery(TimeStampedModel):
    multi_value_report = models.ForeignKey(MultiValueReport, on_delete=models.CASCADE)
    name = models.CharField(max_length=200)
    report_type = models.ForeignKey(ReportType, null=True, blank=True, on_delete=models.PROTECT)
    query = models.JSONField(null=True, blank=True)

    def __str__(self):
        return self.name


class MultiValueReportColumn(TimeStampedModel):
    class WidthType(models.IntegerChoices):
        PERCENTAGE = 0, '%'
        PIXELS = 1, 'Pixels'

    multi_value_report = models.ForeignKey(MultiValueReport, on_delete=models.CASCADE)
    column = models.PositiveSmallIntegerField()
    width_type = models.PositiveSmallIntegerField(choices=WidthType.choices, default=WidthType.PERCENTAGE)
    width = models.PositiveSmallIntegerField()

    def get_td_style(self):
        if self.width_type == self.WidthType.PERCENTAGE:
            return f'width: {self.width}%'
        elif self.width_type == self.WidthType.PIXELS:
            return f'width: {self.width}px'
        return ''


class MultiValueReportCell(TimeStampedModel):
    class MultiValueType(models.IntegerChoices):
        STATIC_TEXT = 0, 'Static Text'
        COUNT = 1, 'Count'
        SUM = 2, 'Sum'
        PERCENT = 4, 'Percent'
        PERCENT_FROM_COUNT = 5, 'Percent from Count'
        AVERAGE_SUM_FROM_COUNT = 6, 'Average Sum from Count'
        AVERAGE_SUM_OVER_TIME = 7, 'Average Sum over Time'
        AVERAGE_SUM_OVER_TIME_EXCLUDING_WEEKENDS = 8, 'Average Sum over Time (Excluding Weekends)'
        EQUATION = 9, 'Equation'

        @classmethod
        def is_percentage(cls, value):
            return value in {
                cls.PERCENT,
                cls.PERCENT_FROM_COUNT,
            }

    multi_value_report = models.ForeignKey(MultiValueReport, on_delete=models.CASCADE)
    row = models.PositiveSmallIntegerField()
    column = models.PositiveSmallIntegerField()
    col_span = models.PositiveSmallIntegerField(default=1)
    row_span = models.PositiveSmallIntegerField(default=1)
    multi_value_type = models.IntegerField(choices=MultiValueType.choices, default=MultiValueType.STATIC_TEXT)
    text = models.TextField(blank=True, null=True)
    multi_cell_style = models.ForeignKey('MultiCellStyle', on_delete=models.SET_NULL, null=True, blank=True)
    report_type = models.ForeignKey(ReportType, null=True, blank=True, on_delete=models.PROTECT)

    field = models.CharField(max_length=200, blank=True, null=True)  # denominator
    numerator = models.CharField(max_length=200, blank=True, null=True)
    prefix = models.CharField(max_length=64, blank=True, null=True)
    decimal_places = models.IntegerField(default=0)

    show_breakdown = models.BooleanField(default=False)
    breakdown_fields = models.JSONField(null=True, blank=True)

    average_scale = models.PositiveSmallIntegerField(choices=ANNOTATION_VALUE_CHOICES, blank=True, null=True)
    average_start_period = models.PositiveSmallIntegerField(blank=True, null=True)
    average_end_period = models.PositiveSmallIntegerField(blank=True, null=True)
    label = models.CharField(max_length=256, blank=True, null=True)

    multi_value_held_query = models.ForeignKey('MultiValueHeldQuery', on_delete=models.SET_NULL, null=True, blank=True)

    query_data = models.JSONField(null=True, blank=True)
    extra_query_data = models.JSONField(null=True, blank=True)  # used for single value Numerator

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=['multi_value_report', 'row', 'column'], name='multi_value_report_cell_unique'
            ),
        ]

    def get_base_model(self):
        if self.report_type is None:
            return None
        return self.report_type.content_type.model_class()


class CalendarReport(Report):
    VIEW_TYPE_CODES = {
        CALENDAR_VIEW_TYPE_MONTH: 'dayGridMonth',
        CALENDAR_VIEW_TYPE_GRID_WEEK: 'timeGridWeek',
        CALENDAR_VIEW_TYPE_LIST_WEEK: 'listWeek',
        CALENDAR_VIEW_TYPE_DAY: 'timeGridDay',
        CALENDAR_VIEW_TYPE_YEAR: 'year',
    }

    height = models.PositiveSmallIntegerField(default=600)
    view_type = models.PositiveSmallIntegerField(
        choices=CALENDAR_VIEW_TYPE_CHOICES,
        default=CALENDAR_VIEW_TYPE_MONTH,
    )

    def get_view_type_for_calendar(self, view_type=None):
        if view_type is None:
            view_type = self.view_type
        return self.VIEW_TYPE_CODES.get(view_type)

    def show_dashboard_query(self):
        return False

    def dashboard_fields(self, form, dashboard_report):
        choices = [(0, f'Default ({self.view_type})'), *CALENDAR_VIEW_TYPE_CHOICES]
        calendar_view_type = dashboard_report.options.get('calendar_view_type') if dashboard_report.options else None
        form.fields['calendar_view_type'] = ChoiceField(
            choices=choices, required=False, widget=Select2(), initial=calendar_view_type
        )

    def save_extra_dashboard_fields(self, form, dashboard_report):
        options = {'calendar_view_type': form.cleaned_data['calendar_view_type']}
        dashboard_report.options = options


class CalendarReportDescription(TimeStampedModel):
    calendar_report = models.ForeignKey(CalendarReport, on_delete=models.CASCADE)
    name = models.CharField(max_length=200)
    report_type = models.ForeignKey(ReportType, null=True, blank=False, on_delete=models.PROTECT)
    description = models.TextField(blank=True, null=True)
    order = models.PositiveSmallIntegerField()

    def save(self, *args, **kwargs):
        self.set_order_field(extra_filters={'calendar_report': self.calendar_report})
        return super().save(*args, **kwargs)

    def get_base_model(self):
        return self.report_type.content_type.model_class()

    def __str__(self):
        return self.name

    class Meta:
        ordering = ('order',)


class CalendarReportDataSet(TimeStampedModel):
    DISPLAY_TYPE_NAME_AND_DESCRIPTION = 1
    DISPLAY_TYPE_DESCRIPTION_ONLY = 2
    DISPLAY_TYPE_HEADING_ONLY = 3

    DISPLAY_TYPE_CHOICES = (
        (DISPLAY_TYPE_NAME_AND_DESCRIPTION, 'Name and Description'),
        (DISPLAY_TYPE_DESCRIPTION_ONLY, 'Description Only'),
        (DISPLAY_TYPE_HEADING_ONLY, 'Heading Only'),
    )

    END_DATE_TYPE_FIELD = 1
    END_DATE_TYPE_DURATION_FIELD = 2
    END_DATE_TYPE_DURATION_FIXED = 3

    END_DATE_TYPE_CHOICES = (
        (END_DATE_TYPE_FIELD, 'Field'),
        (END_DATE_TYPE_DURATION_FIELD, 'Duration Field'),
        (END_DATE_TYPE_DURATION_FIXED, 'Duration Fixed'),
    )

    calendar_report = models.ForeignKey(CalendarReport, on_delete=models.CASCADE)
    order = models.PositiveSmallIntegerField()
    report_type = models.ForeignKey(ReportType, null=True, blank=False, on_delete=models.PROTECT)
    heading_field = models.CharField(max_length=200, blank=True, null=True)
    name = models.CharField(max_length=200)
    display_type = models.PositiveSmallIntegerField(
        choices=DISPLAY_TYPE_CHOICES,
        default=DISPLAY_TYPE_NAME_AND_DESCRIPTION,
    )
    start_date_field = models.CharField(max_length=200, blank=True, null=True)
    end_date_type = models.PositiveSmallIntegerField(
        choices=END_DATE_TYPE_CHOICES,
        default=END_DATE_TYPE_FIELD,
    )
    end_date_field = models.CharField(max_length=200, blank=True, null=True)
    end_duration_field = models.CharField(max_length=200, blank=True, null=True)
    end_duration = models.PositiveSmallIntegerField(blank=True, null=True)
    background_colour_field = models.CharField(max_length=200, blank=True, null=True)
    link_field = models.CharField(max_length=200, blank=True, null=True)
    calendar_report_description = models.ForeignKey(
        CalendarReportDescription, null=True, blank=False, on_delete=models.CASCADE
    )
    query_data = models.JSONField(null=True, blank=True)

    def save(self, *args, **kwargs):
        self.set_order_field(extra_filters={'calendar_report': self.calendar_report})
        return super().save(*args, **kwargs)

    def get_base_model(self):
        return self.report_type.content_type.model_class()

    class Meta:
        ordering = ('order',)


class CustomReport(Report):
    output_type = models.CharField(max_length=200, blank=True, null=True)
    view_name = models.CharField(max_length=200)
    settings = models.JSONField(null=True, blank=True)


class Dashboard(TimeStampedModel):
    slug = models.SlugField(unique=True)
    slug_alias = models.SlugField(blank=True, null=True)  # used if the slug changes
    name = models.CharField(max_length=200)
    display_option = models.PositiveIntegerField(choices=DISPLAY_OPTION_CHOICES[1:], default=DISPLAY_OPTION_2_PER_ROW)

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        slug_alias = self.slug
        self.make_new_slug(allow_dashes=False, on_edit=True)
        if slug_alias != self.slug:
            self.slug_alias = slug_alias
        return super().save(*args, **kwargs)


class DashboardReport(TimeStampedModel):
    dashboard = models.ForeignKey(Dashboard, on_delete=models.CASCADE)
    order = models.PositiveSmallIntegerField()
    report = models.ForeignKey(Report, on_delete=models.CASCADE)
    top = models.BooleanField(default=False)
    name_override = models.CharField(max_length=200, blank=True, null=True)
    display_option = models.PositiveIntegerField(choices=DISPLAY_OPTION_CHOICES, default=DISPLAY_OPTION_NONE)
    show_versions = models.BooleanField(default=True)
    report_query = models.ForeignKey(ReportQuery, blank=True, null=True, on_delete=models.CASCADE)
    options = models.JSONField(null=True, blank=True)

    def get_class(self, extra_class_name):
        if self.display_option != DISPLAY_OPTION_NONE:
            class_names = DISPLAY_OPTION_CLASSES.get(self.display_option)
        else:
            class_names = DISPLAY_OPTION_CLASSES.get(self.dashboard.display_option)

        if extra_class_name:
            class_names += f' {extra_class_name}'

        return class_names

    def save(self, *args, **kwargs):
        self.set_order_field(extra_filters={'dashboard': self.dashboard, 'top': self.top})
        return super().save(*args, **kwargs)

    class Meta:
        ordering = ['order']
